#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np

from nomad.units import ureg
from nomad.metainfo import (
    MSection, Package, Quantity, SubSection, MEnum, Reference, Datetime, Section)
from nomad.metainfo.metainfo import SectionProxy
from nomad.datamodel.data import EntryData


m_package = Package(name='material_library')


class Chemical(EntryData):
    '''A chemical available in the lab.'''
    m_def = Section(label_quantity='formula')

    chemical_name = Quantity(type=str, a_eln=dict(component='StringEditQuantity'))

    formula = Quantity(
        type=str,
        description='Empirical formula of the chemical (Hill notation).',
        a_eln=dict(component='StringEditQuantity'))

    form = Quantity(
        type=MEnum([
            'solid',
            'powder',
            'liquid',
            'gas']),
        shape=[],
        description='Physical state of the chemical.',
        a_eln=dict(component='RadioEnumEditQuantity'))

    supplier = Quantity(
        type=str,
        description='Supplier of the chemical.',
        a_eln=dict(component='StringEditQuantity'))

    sku_no = Quantity(
        type=str,
        description=' Stock keeping unit of the chemical. (e.g. 244619-50G)',
        a_eln=dict(component='StringEditQuantity'))

    opening_date = Quantity(
        type=Datetime,
        description='Opening date of the chemical.',
        a_eln=dict(component='DateTimeEditQuantity'))

    impurities = Quantity(
        type=str,
        description='''Descriptions of the impurities of the product.''',
        a_eln=dict(component='StringEditQuantity'))

    cas_number = Quantity(
        type=str,
        description='''The CAS number is a unique and unambiguous
                    identifier for a specific substance.''',
        a_eln=dict(component='StringEditQuantity'))

    sds_link = Quantity(
        type=str,
        description='''The corresponding Safety Data Sheet (SDS) of the product.''',
        a_eln=dict(component='FileEditQuantity'),
        a_browser=dict(adaptor='RawFileAdaptor'))

    comments = Quantity(
        type=str,
        description='''Remarks about the chemical beyond the
                    typically collected information. Here you can collect information
                    about your experience using this chemical, for example if it might
                    have been contaminated during an experiment''',
        a_eln=dict(component='RichTextEditQuantity'))

    def normalize(self, archive, logger):
        super().normalize(archive, logger)

        if self.formula:
            archive.metadata.entry_name = self.formula
            if self.chemical_name:
                archive.metadata.entry_name += f' ({self.chemical_name})'


class Maintenance(MSection):
    m_def = Section(label_quantity='datetime')

    datetime = Quantity(
        type=Datetime,
        description='The date and time of the maintenance.',
        a_eln=dict(component='DateTimeEditQuantity'))

    maintainer = Quantity(
        type=MEnum([
            'Markus Scheidgen',
            'Pepe Marquez',
            'Sandor Brockhauser',
            'Sherjeel Shabih',
            'Mohammad Nakhaee',
            'David Sitker']),
        description='Name or alias of the persen that performed the maintenance.',
        a_eln=dict(component='AutocompleteEditQuantity'))

    description = Quantity(
        type=str,
        description='Description of what was done to the instrument.',
        a_eln=dict(component='RichTextEditQuantity'))


class Instrument(EntryData):
    '''An instrument available in the lab.'''

    name = Quantity(
        type=str,
        description='Name of the instrument used',
        a_eln=dict(component='StringEditQuantity'))

    # Rich text material for ELNs
    description = Quantity(
        type=str,
        description=(
            '''Description of the instrument. May include images, free text
            and tables'''),
        a_eln=dict(component='RichTextEditQuantity'))

    user_manual = Quantity(
        type=str,
        description='''Link to pdf of the user manual''',
        a_eln=dict(component='FileEditQuantity'),
        a_browser=dict(adaptor='RawFileAdaptor'))

    certificate = Quantity(
        type=str,
        description='''Link to pdf of the instrument's certificates''',
        a_eln=dict(component='FileEditQuantity'),
        a_browser=dict(adaptor='RawFileAdaptor'))

    maintenance = SubSection(section_def=Maintenance, repeats=True)

    def normalize(self, archive, logger):
        super().normalize(archive, logger)

        if self.name:
            archive.metadata.entry_name = self.name


class Process(MSection):
    ''' Any physical process applied to the sample. '''
    operator = Quantity(
        type=MEnum([
            'Markus Scheidgen',
            'Pepe Marquez',
            'Sandor Brockhauser',
            'Sherjeel Shabih',
            'Mohammad Nakhaee',
            'David Sitker']),
        shape=[],
        description='Name or alias of the process operator.',
        a_eln=dict(
            component='AutocompleteEditQuantity'))

    datetime = Quantity(
        type=Datetime,
        description='Finishing date and time of the process.',
        a_eln=dict(component='DateTimeEditQuantity'))

    instrument = Quantity(
        type=Reference(Instrument.m_def),
        descriptions='The instrument used for this process.',
        a_eln=dict(component='ReferenceEditQuantity'))

    chemicals = Quantity(
        type=Reference(Chemical.m_def),
        shape=['*'],
        descriptions='The chemicals used in this process',
        a_eln=dict(component='ReferenceEditQuantity'))

    creates_layer = Quantity(
        type=bool,
        a_eln=dict(component='BoolEditQuantity'))

    comments = Quantity(
        type=str,
        description='''Remarks about the process that cannot be seen from the data.
                    Might include rich text, images and potentially tables''',
        a_eln=dict(component='RichTextEditQuantity'))

    def normalize(self, archive, logger):
        if self.creates_layer:
            logger.debug('create layer if necessary')
            layer_exists = False
            for layer in archive.data.layers:
                if layer.layer_origin is None:
                    continue
                if layer.layer_origin.m_resolved() == self:
                    layer_exists = True

            if not layer_exists:
                logger.debug('create layer')
                archive.data.layers.append(Layer(
                    layer_origin=self,
                    layer_creation_datetime=self.datetime))


class PVDEvaporation(Process):
    '''The physical vapor deposition (PVD) of a layer by evaporation.'''

    def derive_n_values(self):
        if self.process_time is not None:
            return len(self.process_time)
        if self.set_substrate_temperature is not None:
            return len(self.set_substrate_temperature)
        else:
            return 0

    n_values = Quantity(
        type=int, derived=derive_n_values,
        description='Number of registered time values')

    process_time = Quantity(
        type=np.dtype(np.float64), unit='seconds', shape=['n_values'],
        description='The temperature set in the substrate heater')

    set_substrate_temperature = Quantity(
        type=np.dtype(np.float64), unit='kelvin', shape=['n_values'],
        description='The temperature set in the substrate heater',
        a_plot={
            'x': 'process_time', 'y': 'set_substrate_temperature'
        })

    substrate_temperature = Quantity(
        type=np.dtype(np.float64), unit='kelvin', shape=['n_values'],
        description='The temperature measured in the substrate holder',
        a_plot={
            'x': 'process_time', 'y': 'substrate_temperature'
        })

    chamber_pressure = Quantity(
        type=np.dtype(np.float64), unit='pascal', shape=['n_values'],
        description='Data array of the values of the pressure during the process',
        a_plot={
            'x': 'process_time', 'y': 'chamber_pressure'
        })

    number_crucibles = Quantity(
        type=np.dtype(np.int64),
        description='Number of crucibles active in the chamber',
        a_eln=dict(component='NumberEditQuantity', props=dict(maxValue=4, minValue=0)))

    data_file = Quantity(
        type=str,
        a_eln=dict(component='FileEditQuantity'),
        a_browser=dict(adaptor='RawFileAdaptor'))

    def normalize(self, archive, logger):
        super().normalize(archive, logger)

        from nomad.datamodel.metainfo.material_library.PvdPImporter import Importer
        importer = Importer()
        if (self.data_file):
            with archive.m_context.raw_file(self.data_file) as f:
                self.process_time, substrate_temperature, \
                    set_substrate_temperature, chamber_pressure = importer.read(f)
                self.substrate_temperature = substrate_temperature * ureg.degC
                self.set_substrate_temperature = set_substrate_temperature * ureg.degC
                self.chamber_pressure = chamber_pressure * ureg('mbar')
                # self.n_values = self.n_values


class Targets(MSection):
    '''Targets available in the lab for PLD or Sputtering deposition.'''

    target_name = Quantity(
        type=str,
        description='''Description of the target including information of its
                     chemistry nature . Examples: *Al doped ZnO* or simply *BaZrO3*''')

    elements_in_target = Quantity(type=str, shape=["*"])

    target_diameter = Quantity(type=np.dtype(np.float64), unit='meter')

    supplier = Quantity(
        type=str,
        description='Supplier of the target.')

    opening_date = Quantity(
        type=Datetime,
        description='Opening date of the chemical.')

    impurities = Quantity(
        type=str,
        description='''Descriptions of the impurities of the product.''')

    cas_number = Quantity(
        type=str,
        description='''The CAS number is a unique and unambiguous
                    identifier for a specific substance.''')

    sds_link = Quantity(
        type=str,
        description='''A link to the corresponding Safety Data Sheet (SDS)
                    of the product.''')

    comments = Quantity(
        type=str,
        description='''Remarks about the target beyond the
                    typically collected information in the otehr metadata fields.
                    Here you can collect information about your experience using
                    this target, or additional observations''')


class PLDDeposition(Process):
    '''The deposition process of a layer by Pulsed Laser Deposition (PLD) method.'''

    targets = SubSection(section_def=Targets, repeats=True)


class EbeamEvaporation(Process):
    '''The physical vapor deposition (PVD) of a layer by e-beam evaporation.'''


class HotPlateAnnealing(Process):

    hotplate_temperature = Quantity(
        type=np.dtype(np.float64), unit='kelvin', shape=[],
        description='The temperature set for the hot plate.',
        a_eln=dict(component='NumberEditQuantity'))

    annealing_time = Quantity(
        type=np.dtype(np.float64), unit='seconds', shape=[],
        description='Time of the sample on the hot plate.',
        a_eln=dict(component='NumberEditQuantity'), props=dict(minValue=0))

    relative_humidity = Quantity(
        type=np.dtype(np.float64),  # unit='',
        description=('''
            Relative humidity of the atmosphere in which the experiment was performed.
        '''),
        a_eln=dict(component='NumberEditQuantity', props=dict(minValue=0, maxValue=1)))

    instrument_atmosphere = Quantity(
        type=MEnum([
            'humidity chamber',
            'glove box',
            'ambient']),
        shape=[],
        description='Location or atmosphere in which the process was conducted.',
        a_eln=dict(
            label='Process atmosphere',
            component='RadioEnumEditQuantity'))


class TubeFurnaceAnnealing(Process):
    pass


class RTPAnnealing(Process):
    pass


class SpinCoating(Process):
    pass


class ChemicalBathDeposition(Process):
    pass


class Processes(MSection):
    '''
    Experiment event which  generally change the sample or a new component is added to it.
    For example, in the context of thin films, cleaning the substrate or
    the deposition of a new layer by evaporation are `processes`.
    '''
    m_def = Section(a_eln=dict())

    pvd_evaporation = SubSection(section_def=PVDEvaporation)
    pld_deposition = SubSection(section_def=PLDDeposition)
    ebeam_evaporation = SubSection(section_def=EbeamEvaporation)
    hotplate_annealing = SubSection(section_def=HotPlateAnnealing)
    tubefurnace_annealing = SubSection(section_def=TubeFurnaceAnnealing)
    rtp_annealing = SubSection(section_def=RTPAnnealing)
    spin_coating = SubSection(section_def=SpinCoating)
    chemical_bath_deposition = SubSection(section_def=ChemicalBathDeposition)


class Measurement(MSection):
    '''
    Any measurement performed on the sample.
    '''
    instrument = Quantity(
        type=Reference(Instrument.m_def),
        descriptions='The instrument used for this measurement.',
        a_eln=dict(component='ReferenceEditQuantity'))

    comments = Quantity(
        type=str,
        description='Remarks about the process that cannot be seen from the data.',
        a_eln=dict(component='RichTextEditQuantity'))


class XrayFluorescence(Measurement):
    '''
    X-ray fluorescence is a technique typically used to obtain information about the
    chemical composition of a sample.
    '''


class XrayDiffraction(Measurement):
    '''
    X-ray diffraction is a technique typically used to characterize the structural
    properties of crystalline materials. The data contains `two_theta` values of the scan
    the corresponding counts collected for each channel
    '''

    start_position_x = Quantity(
        type=np.dtype(np.float64), shape=['*'], unit='meter',
        description='''Length from the left substrate margin to the spot of
                    the first scan. The sample should be oriented with the
                    label at the bottom left corner''',
        a_eln=dict(
            component='NumberEditQuantity',
            label='Length from the left substrate margin'))

    start_position_y = Quantity(
        type=np.dtype(np.float64), shape=['*'], unit='meter',
        description='''Length from the bottom substrate margin to the spot of
                    the first scan. The sample should be oriented with the
                    label at the bottom left corner.''',
        a_eln=dict(
            component='NumberEditQuantity',
            label='Length from the bottom substrate margin'))

    def derive_n_values(self):
        if self.intensity is not None:
            return len(self.intensity)
        if self.two_theta is not None:
            return len(self.two_theta)
        else:
            return 0

    n_values = Quantity(type=int, derived=derive_n_values)

    two_theta = Quantity(
        type=np.dtype(np.float64), shape=['n_values'],
        unit='radian',
        description='The 2-theta range of the difractogram',
        a_plot={
            'x': 'two_theta', 'y': 'intensity'
        })

    q_vector = Quantity(
        type=np.dtype(np.float64), shape=['n_values'],
        unit='meter**(-1)',
        description='The scattering vector *Q* of the difractogram',
        a_plot={
            'x': 'q_vector', 'y': 'intensity'
        })

    intensity = Quantity(
        type=np.dtype(np.float64), shape=['n_values'],
        description='The count at each 2-theta value, dimensionless',
        a_plot={
            'x': 'two_theta', 'y': 'intensity'
        })

    omega = Quantity(
        type=np.dtype(np.float64), shape=['n_values'],
        unit='radian',
        description='The omega range of the difractogram')

    xray_tube_current = Quantity(
        type=np.dtype(np.float64),
        unit='A',
        description='Current of the X-ray tube',
        a_eln=dict(
            component='NumberEditQuantity',
            label='Current of the X-ray tube',
            props=dict(minValue=1e4, maxValue=4e4)))

    xray_tube_voltage = Quantity(
        type=np.dtype(np.float64),
        unit='V',
        description='Voltage of the X-ray tube',
        a_eln=dict(
            component='NumberEditQuantity',
            label='Voltage of the X-ray tube',
            props=dict(minValue=1e4, maxValue=4e4)))

    kalpha_one = Quantity(
        type=np.dtype(np.float64),
        unit='angstrom',
        description='Wavelength of the Kα1 line')

    kalpha_two = Quantity(
        type=np.dtype(np.float64),
        unit='angstrom',
        description='Wavelength of the Kα2 line')

    ratio_kalphatwo_kalphaone = Quantity(
        type=np.dtype(np.float64),
        description='Kα2/Kα1 intensity ratio')

    kbeta = Quantity(
        type=np.dtype(np.float64),
        unit='angstrom',
        description='Wavelength of the Kβ line')

    scan_axis = Quantity(
        type=str,
        description='Axis scanned')

    integration_time = Quantity(
        type=np.dtype(np.float64),
        shape=['*'],
        description='Integration time per channel')

    data_file = Quantity(
        type=str,
        a_eln=dict(component='FileEditQuantity'))

    def normalize(self, archive, logger):
        import xrdtools
        if (self.data_file):
            with archive.m_context.raw_file(self.data_file) as f:
                xrdml_dict = xrdtools.read_xrdml(f.name)
                self.intensity = xrdml_dict['data']
                self.two_theta = xrdml_dict['x'] * ureg('degree')
                self.omega = xrdml_dict['Omega'] * ureg('degree')
                self.kalpha_one = xrdml_dict['kAlpha1']
                self.kalpha_two = xrdml_dict['kAlpha2']
                self.ratio_kalphatwo_kalphaone = xrdml_dict['kAlphaRatio']
                self.kbeta = xrdml_dict['kBeta']
                self.scan_axis = xrdml_dict['scanAxis']
                self.integration_time = xrdml_dict['time']
                self.q_vector = (4 * np.pi / self.kalpha_one) * np.sin((self.two_theta))


class RamanSpectroscopy(Measurement): pass


class Resistivity(Measurement): pass


class SolarSimulator(Measurement): pass


class CapacitanceVoltage(Measurement): pass


class EQE(Measurement): pass


class SteadyStatePL(Measurement): pass


class PLImaging(Measurement): pass


class TimeResolvedPL(Measurement): pass


class UVVisNIRImaging(Measurement): pass


class UVVisNIRSpectroscopy(Measurement): pass


class TerahertzSpectroscopy(Measurement): pass


class Measurements(MSection):
    '''
    Experimental procedure in which a sample gets characterized
    by a technique. For example, a measurment by X-ray diffraction to characterize
    the structural properties of an specimen or X-ray fluorescence
    to characterize its composition.
    '''
    m_def = Section(a_eln=dict())

    xray_diffraction = SubSection(section_def=XrayDiffraction, repeats=True)
    xray_fluorescence = SubSection(section_def=XrayFluorescence, repeats=True)
    raman_spectroscopy = SubSection(section_def=RamanSpectroscopy, repeats=True)
    resistivity = SubSection(section_def=Resistivity, repeats=True)
    solar_simulator = SubSection(section_def=SolarSimulator, repeats=True)
    capacitance_voltage = SubSection(section_def=CapacitanceVoltage, repeats=True)
    eqe = SubSection(section_def=EQE, repeats=True)
    steady_state_pl = SubSection(section_def=SteadyStatePL, repeats=True)
    pl_imaging = SubSection(section_def=PLImaging, repeats=True)
    time_resolved_pl = SubSection(section_def=TimeResolvedPL, repeats=True)
    uv_vis_nir_imaging = SubSection(section_def=UVVisNIRImaging, repeats=True)
    uv_vis_spectroscopy = SubSection(section_def=UVVisNIRSpectroscopy, repeats=True)
    terahertz_spectroscopy = SubSection(section_def=TerahertzSpectroscopy, repeats=True)


class DerivedData(MSection):
    '''
    Additional information that is gained from already existing data i.e. from `processes`,
     `measurements`  and/or other `derived_data`. Examples for these are smoothing of data,
    background subtraction, calculations (e.g. transmittance from raw spectra,
    bandgap from the absorption coefficient) or simulations.
    '''


class PhysicalProperties(MSection):
    '''
    It describes physical properties of the layer, like its dimensions in the *x* and *y*
    directions or its thickness.
    '''
    m_def = Section(a_eln=dict())

    x_length = Quantity(
        type=np.dtype(np.float64), unit='meter', shape=[],
        description='Dimension of the layer in the *x*-direction',
        a_eln=dict(component='NumberEditQuantity'))

    y_length = Quantity(
        type=np.dtype(np.float64), unit='meter', shape=[],
        description='Dimension of the layer in the *y*-direction',
        a_eln=dict(component='NumberEditQuantity'))

    thickness = Quantity(
        type=np.dtype(np.float64), unit='meter', shape=["*"],
        description='Thickness of the layer',
        a_eln=dict(component='NumberEditQuantity'))


class CompositionalProperties(MSection):
    '''
    It describes the chemical properties of the layer, like the *elements* contained
    in the layer or the *compounds* if they are known. These properties can be filled with
    knowledge obtained from `measurements`.
    '''
    elements = Quantity(type=str, shape=["*"])
    chemical_formula = Quantity(type=str)


class StructuralProperties(MSection): pass


class OptoelectronicProperties(MSection): pass


class Layer(MSection):
    m_def = Section(a_eln=dict())

    layer_type = Quantity(
        type=MEnum([
            'substrate',
            'absorber layer',
            'metal contact',
            'buffer layer',
            'p-type contact',
            'n-type contact']),
        shape=[],
        description='type of the layer',
        a_eln=dict(component='EnumEditQuantity'))

    layer_creation_datetime = Quantity(
        type=Datetime,
        description='Creation date of the layer.',
        a_eln=dict(
            label='Creation date and time of the layer.',
            component='DateTimeEditQuantity'))

    layer_position = Quantity(
        type=np.dtype(np.int64),
        description='''Position of the layer within
        the stack counting from substrate = 0; not necessarily gapless''',
        a_eln=dict(component='NumberEditQuantity'))

    layer_origin = Quantity(
        type=Reference(Process.m_def),
        description='A link to the `process` where the layer was created.')

    x_length = Quantity(
        type=np.dtype(np.float64), unit='meter', shape=[],
        description='Dimension of the layer in the *x*-direction',
        a_eln=dict(component='NumberEditQuantity'))

    y_length = Quantity(
        type=np.dtype(np.float64), unit='meter', shape=[],
        description='Dimension of the layer in the *y*-direction',
        a_eln=dict(component='NumberEditQuantity'))

    thickness = Quantity(
        type=np.dtype(np.float64), unit='meter', shape=["*"],
        description='Thickness of the layer',
        a_eln=dict(component='NumberEditQuantity'))

    elements = Quantity(type=str, shape=["*"], a_eln=dict(component='StringEditQuantity'))
    chemical_formula = Quantity(type=str, a_eln=dict(component='StringEditQuantity'))

    # physical_properties = SubSection(section_def=PhysicalProperties)
    # compositional_properties = SubSection(section_def=CompositionalProperties)
    # structural_properties = SubSection(section_def=StructuralProperties)
    # optoelectronic_properties = SubSection(section_def=OptoelectronicProperties)


class Projects(MSection):
    project_acronym = Quantity(
        type=str, description='Acronim or short name of the project.')

    project_long_name = Quantity(
        type=str, description='Long name of the prioject.')

    # Might cotain images and free text in ELNs features
    project_description = Quantity(
        type=str, description='A description of the project activities.')

    project_website = Quantity(
        type=str, description='Project website.')

    funding_agency = Quantity(
        type=str, description='The physical state of the sample.')

    gran_agreement_id = Quantity(
        type=str, description='ID of the contract of that is funding this project.')


class Sample(EntryData):
    '''
    Sample in which one or various properties
    can vary across the area of the substrate,
    thus containing subsamples whithin.
    An example would be a subtrate with a deposited film on top in which
    the chemical compostion of the film varies in the x and y directions.
    '''

    sample_owner = Quantity(
        type=MEnum([
            'Markus Scheidgen',
            'Pepe Marquez',
            'Sandor Brockhauser',
            'Sherjeel Shabih',
            'Mohammad Nakhaee',
            'David Sitker']),
        shape=[],
        description='Name or alias of the process operator.',
        a_eln=dict(component='AutocompleteEditQuantity'))

    sample_id = Quantity(
        type=str,
        description='Full sample id.',
        a_eln=dict(component='StringEditQuantity'))

    sample_name = Quantity(
        type=str,
        description='Short name of the sample (the id on the substrate), e.g. `4001-8`.',
        a_eln=dict(component='StringEditQuantity'))

    creation_datetime = Quantity(
        type=Datetime,
        description='Creation date of the sample.',
        a_eln=dict(component='DateTimeEditQuantity'))

    institute = Quantity(
        type=str,
        description='Alias/short name of the home institute of the owner, i.e. `HZB`.',
        a_eln=dict(component='StringEditQuantity'))

    description = Quantity(
        type=str, a_eln=dict(component='RichTextEditQuantity'))

    processes = SubSection(section_def=Processes)
    measurements = SubSection(section_def=Measurements)
    # derived_data = SubSection(section_def=DerivedData, repeats=True)
    layers = SubSection(section_def=Layer, repeats=True)
    # projects = SubSection(section_def=Projects, repeats=True)


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
This module describes all configurable parameters for the nomad python code. The
configuration is used for all executed python code including API, worker, CLI, and other
scripts. To use the configuration in your own scripts or new modules, simply import
this module.

All parameters are structured into objects for two reasons. First, to have
categories. Second, to allow runtime manipulation that is not effected
by python import logic. The categories are choosen along infrastructure components:
``mongo``, ``elastic``, etc.

This module also provides utilities to read the configuration from environment variables
and .yaml files. This is done automatically on import. The precedence is env over .yaml
over defaults.

.. autoclass:: nomad.config.NomadConfig
.. autofunction:: nomad.config.load_config
'''

import logging
import os
import os.path
import yaml
import warnings
from typing import Dict, Any

try:
    from nomad import gitinfo
except ImportError:
    git_root = os.path.join(os.path.dirname(__file__), '..')
    cwd = os.getcwd()
    os.chdir(git_root)
    os.system('./gitinfo.sh')
    os.chdir(cwd)

    from nomad import gitinfo


warnings.filterwarnings('ignore', message='numpy.dtype size changed')
warnings.filterwarnings('ignore', message='numpy.ufunc size changed')


class NomadConfig(dict):
    '''
    A class for configuration categories. It is a dict subclass that uses attributes as
    key/value pairs.
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def customize(self, custom_settings: Dict[str, Any]) -> 'NomadConfig':
        '''
        Returns a new NomadConfig object, created by taking a copy of the current config and
        updating it with the settings defined in `custom_settings`. The `custom_settings` dict
        must not contain any new keys (keys not defined in this NomadConfig). If it does,
        an exception will be raised.
        '''
        rv = NomadConfig(**self)
        if custom_settings:
            for k, v in custom_settings.items():
                assert k in rv, f'Invalid setting: {k}'
                rv[k] = v
        return rv


CELERY_WORKER_ROUTING = 'worker'
CELERY_QUEUE_ROUTING = 'queue'

rabbitmq = NomadConfig(
    host='localhost',
    user='rabbitmq',
    password='rabbitmq'
)


def rabbitmq_url():
    return 'pyamqp://%s:%s@%s//' % (rabbitmq.user, rabbitmq.password, rabbitmq.host)


celery = NomadConfig(
    max_memory=64e6,  # 64 GB
    timeout=1800,  # 1/2 h
    acks_late=False,
    routing=CELERY_QUEUE_ROUTING,
    priorities={
        'Upload.process_upload': 5,
        'Upload.delete_upload': 9,
        'Upload.publish_upload': 10
    }
)

fs = NomadConfig(
    tmp='.volumes/fs/tmp',
    staging='.volumes/fs/staging',
    staging_external=None,
    public='.volumes/fs/public',
    public_external=None,
    local_tmp='/tmp',
    prefix_size=2,
    archive_version_suffix='v1',
    working_directory=os.getcwd()
)

try:
    fs.staging_external = os.path.abspath(fs.staging)
    fs.public_external = os.path.abspath(fs.public)
except Exception:
    pass

elastic = NomadConfig(
    host='localhost',
    port=9200,
    timeout=60,
    bulk_timeout=600,
    bulk_size=1000,
    entries_per_material_cap=1000,
    entries_index='nomad_entries_v1',
    materials_index='nomad_materials_v1',
)

keycloak = NomadConfig(
    public_server_url=None,
    server_url='https://nomad-lab.eu/fairdi/keycloak/auth/',
    realm_name='fairdi_nomad_prod',
    username='admin',
    password='password',
    client_id='nomad_public',
    client_secret=None,
    oasis=False)

mongo = NomadConfig(
    host='localhost',
    port=27017,
    db_name='nomad_v1'
)

logstash = NomadConfig(
    enabled=False,
    host='localhost',
    tcp_port='5000',
    level=logging.DEBUG
)

services = NomadConfig(
    api_host='localhost',
    api_port=8000,
    api_base_path='/fairdi/nomad/latest',
    api_secret='defaultApiSecret',
    api_chaos=0,
    admin_user_id='00000000-0000-0000-0000-000000000000',
    not_processed_value='not processed',
    unavailable_value='unavailable',
    https=False,
    https_upload=False,
    upload_limit=10,
    force_raw_file_decoding=False,
    download_scan_size=500,
    download_scan_timeout=u'30m'
)

oasis = NomadConfig(
    central_nomad_api_url='https://nomad-lab.eu/prod/v1/api',
    central_nomad_deployment_id='nomad-lab.eu/prod/v1',
    allowed_users=None  # a list of usernames or user account emails
)

tests = NomadConfig(
    default_timeout=30
)


def api_url(ssl: bool = True, api: str = 'api', api_host: str = None, api_port: int = None):
    '''
    Returns the url of the current running nomad API. This is for server-side use.
    This is not the NOMAD url to use as a client, use `nomad.config.client.url` instead.
    '''
    if api_port is None:
        api_port = services.api_port
    if api_host is None:
        api_host = services.api_host
    protocol = 'https' if services.https and ssl else 'http'
    host_and_port = api_host
    if api_port not in [80, 443]:
        host_and_port += ':' + str(api_port)
    base_path = services.api_base_path.strip('/')
    return f'{protocol}://{host_and_port}/{base_path}/{api}'


def gui_url(page: str = None):
    base = api_url(True)[:-3]
    if base.endswith('/'):
        base = base[:-1]

    if page is not None:
        return '%s/gui/%s' % (base, page)

    return '%s/gui' % base


def _check_config():
    """Used to check that the current configuration is valid. Should only be
    called once after the final config is loaded.

    Raises:
        AssertionError: if there is a contradiction or invalid values in the
            config file settings.
    """
    # The AFLOW symmetry information is checked once on import
    proto_symmetry_tolerance = normalize.prototype_symmetry_tolerance
    symmetry_tolerance = normalize.symmetry_tolerance
    if proto_symmetry_tolerance != symmetry_tolerance:
        raise AssertionError(
            "The AFLOW prototype information is outdated due to changed tolerance "
            "for symmetry detection. Please update the AFLOW prototype information "
            "by running the CLI command 'nomad admin ops prototype-update "
            "--matches-only'."
        )

    if normalize.springer_db_path and not os.path.exists(normalize.springer_db_path):
        normalize.springer_db_path = None

    if keycloak.public_server_url is None:
        keycloak.public_server_url = keycloak.server_url


mail = NomadConfig(
    enabled=False,
    with_login=False,
    host='',
    port=8995,
    user='',
    password='',
    from_address='support@nomad-lab.eu',
    cc_address='support@nomad-lab.eu'
)

normalize = NomadConfig(
    # The system size limit for running the dimensionality analysis. For very
    # large systems the dimensionality analysis will get too expensive.
    system_classification_with_clusters_threshold=64,
    # Symmetry tolerance controls the precision used by spglib in order to find
    # symmetries. The atoms are allowed to move 1/2*symmetry_tolerance from
    # their symmetry positions in order for spglib to still detect symmetries.
    # The unit is angstroms. The value of 0.1 is used e.g. by Materials Project
    # according to
    # https://pymatgen.org/pymatgen.symmetry.analyzer.html#pymatgen.symmetry.analyzer.SpacegroupAnalyzer
    symmetry_tolerance=0.1,
    # The symmetry tolerance used in aflow prototype matching. Should only be
    # changed before re-running the prototype detection.
    prototype_symmetry_tolerance=0.1,
    # Maximum number of atoms in the single cell of a 2D material for it to be
    # considered valid.
    max_2d_single_cell_size=7,
    # The distance tolerance between atoms for grouping them into the same
    # cluster. Used in detecting system type.
    cluster_threshold=2.5,
    # Defines the "bin size" for rounding cell angles for the material hash
    angle_rounding=float(10.0),  # unit: degree
    # The threshold for a system to be considered "flat". Used e.g. when
    # determining if a 2D structure is purely 2-dimensional to allow extra rigid
    # transformations that are improper in 3D but proper in 2D.
    flat_dim_threshold=0.1,
    # The threshold for point equality in k-space. Unit: 1/m.
    k_space_precision=150e6,
    # The energy threshold for how much a band can be on top or below the fermi
    # level in order to still detect a gap. Unit: Joule.
    band_structure_energy_tolerance=8.01088e-21,  # 0.05 eV
    springer_db_path=os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'normalizing/data/springer.msg'
    )
)

paths = NomadConfig(
    similarity="",
)

client = NomadConfig(
    user='leonard.hofstadter@nomad-fairdi.tests.de',
    password='password',
    url='http://nomad-lab.eu/prod/v1/api'
)

datacite = NomadConfig(
    mds_host='https://mds.datacite.org',
    enabled=False,
    prefix='10.17172',
    user='*',
    password='*'
)

meta = NomadConfig(
    version='1.1.0',
    commit=gitinfo.commit,
    deployment='devel',
    label=None,
    default_domain='dft',
    service='unknown nomad service',
    name='novel materials discovery (NOMAD)',
    description='A FAIR data sharing platform for materials science data',
    homepage='https://nomad-lab.eu',
    source_url='https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-FAIR',
    maintainer_email='markus.scheidgen@physik.hu-berlin.de',
    deployment_id='nomad-lab.eu/prod/v1',
    beta=None
)

gitlab = NomadConfig(
    private_token='not set'
)

reprocess = NomadConfig(
    # Configures standard behaviour when reprocessing.
    # Note, the settings only matter for published uploads and entries. For uploads in
    # staging, we always reparse, add newfound entries, and delete unmatched entries.
    rematch_published=True,
    reprocess_existing_entries=True,
    use_original_parser=False,
    add_matched_entries_to_published=True,
    delete_unmatched_published_entries=False,
    index_invidiual_entries=False
)

process = NomadConfig(
    index_materials=True,
    reuse_parser=True,
    metadata_file_name='nomad',
    metadata_file_extensions=('json', 'yaml', 'yml')
)

bundle_import = NomadConfig(
    # Basic settings
    allow_bundles_from_oasis=True,  # If oasis admins can "push" bundles to this NOMAD deployment
    allow_unpublished_bundles_from_oasis=False,  # If oasis admins can "push" bundles of unpublished uploads
    required_nomad_version='1.0.0',  # Minimum  nomad version of bundles required for import

    default_settings=NomadConfig(
        # Default settings for the import_bundle process.
        # Note, admins, and only admins, can override these settings when importing a bundle.
        # This means that if oasis admins pushes bundles to this NOMAD deployment, these
        # default settings will be applied.
        include_raw_files=True,
        include_archive_files=False,
        include_datasets=True,
        include_bundle_info=True,  # Keeps the bundle_info.json file (not necessary but nice to have)
        keep_original_timestamps=False,  # If all time stamps (create time, publish time etc) should be imported from the bundle
        set_from_oasis=True,  # If the from_oasis flag and oasis_deployment_id should be set
        delete_upload_on_fail=False,  # If False, it is just removed from the ES index on failure
        delete_bundle_when_done=True,  # Deletes the source bundle when done (regardless of success)
        also_delete_bundle_parent_folder=True,  # Also deletes the parent folder, if it is empty.
        trigger_processing=True,  # If the upload should be processed when the import is done.

        # When importing with trigger_processing=True, the settings below control the
        # initial processing behaviour (see the config for `reprocess` for more info).
        rematch_published=True,
        reprocess_existing_entries=True,
        use_original_parser=False,
        add_matched_entries_to_published=True,
        delete_unmatched_published_entries=False
    )
)

north = NomadConfig(
    hub_ip_connect='172.17.0.1',  # Set this to host.docker.internal on windows/macos.
    hub_connect_url=None,
    hub_ip='0.0.0.0',
    docker_network=None,
    hub_host='localhost',
    hub_port=9000,
    shared_fs='.volumes/fs/north/shared',
    users_fs='.volumes/fs/north/users',
    jupyterhub_crypt_key=None,
    windows=True  # enable windows (as in windows the OS) hacks
)


def north_url(ssl: bool = True):
    return api_url(ssl=ssl, api='north', api_host=north.hub_host, api_port=north.hub_port)


auxfile_cutoff = 100
parser_matching_size = 150 * 80  # 150 lines of 80 ASCII characters per line
console_log_level = logging.WARNING
max_upload_size = 32 * (1024 ** 3)
raw_file_strip_cutoff = 1000
max_entry_download = 500000
encyclopedia_base = "https://nomad-lab.eu/prod/rae/encyclopedia/#"
aitoolkit_enabled = False
use_empty_parsers = False


def normalize_loglevel(value, default_level=logging.INFO):
    plain_value = value
    if plain_value is None:
        return default_level
    else:
        try:
            return int(plain_value)
        except ValueError:
            return getattr(logging, plain_value)


_transformations = {
    'console_log_level': normalize_loglevel,
    'logstash_level': normalize_loglevel
}


# use std python logger, since logging is not configured while loading configuration
logger = logging.getLogger(__name__)


def _apply(key, value, raise_error: bool = True) -> None:
    '''
    Changes the config according to given key and value. The first part of a key
    (with ``_`` as a separator) is interpreted as a group of settings. E.g. ``fs_staging``
    leading to ``config.fs.staging``.
    '''
    full_key = key
    try:
        group_key, config_key = full_key.split('_', 1)
    except Exception:
        if raise_error:
            logger.error(f'config key does not exist: {full_key}')
        return

    current = globals()

    if group_key not in current:
        if key not in current:
            if raise_error:
                logger.error(f'config key does not exist: {full_key}')
            return
    else:
        current = current[group_key]
        if not isinstance(current, NomadConfig):
            if raise_error:
                logger.error(f'config key does not exist: {full_key}')
            return

        if config_key not in current:
            if raise_error:
                logger.error(f'config key does not exist: {full_key}')
            return

        key = config_key

    try:
        current_value = current[key]
        if current_value is not None and not isinstance(value, type(current_value)):
            value = _transformations.get(full_key, type(current_value))(value)

        current[key] = value
        logger.info(f'set config setting {full_key}={value}')
    except Exception as e:
        logger.error(f'cannot set config setting {full_key}={value}: {e}')


def _apply_env_variables():
    kwargs = {
        key[len('NOMAD_'):].lower(): value
        for key, value in os.environ.items()
        if key.startswith('NOMAD_') and key != 'NOMAD_CONFIG'}

    for key, value in kwargs.items():
        _apply(key, value, raise_error=False)


def _apply_nomad_yaml():
    config_file = os.environ.get('NOMAD_CONFIG', 'nomad.yaml')

    if not os.path.exists(config_file):
        return

    with open(config_file, 'r') as stream:
        try:
            config_data = yaml.load(stream, Loader=getattr(yaml, 'FullLoader'))
        except yaml.YAMLError as e:
            logger.error(f'cannot read nomad config: {e}')
            return

    if not config_data:
        return

    for key, value in config_data.items():
        if isinstance(value, dict):
            group_key = key
            for key, value in value.items():
                _apply(f'{group_key}_{key}', value)
        else:
            _apply(key, value)


def load_config():
    '''
    Loads the configuration from nomad.yaml and environment.
    '''
    _apply_nomad_yaml()
    _apply_env_variables()
    _check_config()


load_config()



"""Read files from different format and print it in a standard Nexus format

"""
# Nexus definitions in github: https://github.com/nexusformat/definitions
# to be cloned under os.environ['NEXUS_DEF_PATH']

import os
import xml.etree.ElementTree as ET
import re
import sys
import logging
import textwrap
import h5py

# LOGGING_FORMAT = "%(levelname)s: %(message)s"
# stdout_handler = logging.StreamHandler(sys.stdout)
# stdout_handler.setLevel(logging.DEBUG)
# logging.basicConfig(level=logging.DEBUG, format=LOGGING_FORMAT, handlers=[stdout_handler])
# logger = logging.getLogger()


class NxdlAttributeError(Exception):
    """An exception for throwing an error when an Nxdl attribute is not found."""


# check for NEXUS definitions
def get_nexus_definitions_path():
    """Check NEXUS_DEF_PATH variable.
If it is empty, this function is filling it

"""
    try:
        # either given by sys env
        return os.environ['NEXUS_DEF_PATH']
    except KeyError:
        # or it should be available locally under the dir 'definitions'
        local_dir = os.path.abspath(os.path.dirname(__file__))
        return os.path.join(local_dir, f"..{os.sep}definitions")


def get_nx_class_path(hdf_node):
    """Get the full path of an HDF5 node using nexus classes
in case of a field, end with the field name

"""

    if hdf_node.name == '/':
        return ''
    if isinstance(hdf_node, h5py.Group):
        return get_nx_class_path(
            hdf_node.parent) + '/' + hdf_node.attrs['NX_class']
    if isinstance(hdf_node, h5py.Dataset):
        return get_nx_class_path(
            hdf_node.parent) + '/' + hdf_node.name.split('/')[-1]
    return ''


def get_nxdl_entry(hdf_node):
    """Get the nxdl application definition for an HDF5 node

"""

    entry = hdf_node
    while isinstance(entry,
                     h5py.Dataset) or entry.attrs['NX_class'] != 'NXentry':
        entry = entry.parent
        if entry.name == '/':
            return 'NO NXentry found'
    try:
        nxdef = entry['definition'][()]
        return nxdef.decode()
    except KeyError:
        return 'NO Definition referenced'


def get_nx_class(nxdl_elem):
    """Get the nexus class for a NXDL node

"""
    try:
        return nxdl_elem.attrib['type']
    except KeyError:
        return 'NX_CHAR'


def get_nx_namefit(hdf_name, name):
    """Checks if an HDF5 node name corresponds to a child of the NXDL element
uppercase letters in front can be replaced by arbitraty name, but
uppercase to lowercase match is preferred,
so such match is counted as a measure of the fit

"""
    # count leading capitals
    counting = 0
    while counting < len(name) and name[counting] >= 'A' and name[counting] <= 'Z':
        counting += 1
    # if potential fit
    if counting == len(name) or hdf_name.endswith(name[counting:]):
        # count the matching chars
        fit = 0
        for i in range(max(counting, len(hdf_name))):
            if hdf_name[i].upper() == name[i]:
                fit += 1
            else:
                break
        # accept only full fits as better fits
        if fit == max(counting, len(hdf_name)):
            return fit
        return 0
    # no fit
    return -1


def get_nx_classes():
    """Read base classes from the Nexus definition/base_classes folder

"""
    base_classes_list_files = os.listdir(os.path.join(get_nexus_definitions_path(), 'base_classes'))
    nx_clss = sorted([str(s)[:-9] for s in base_classes_list_files])
    return nx_clss


def get_nx_units():
    """Read unit kinds from the Nexus definition/nxdlTypes.xsd file

"""
    filepath = f"{get_nexus_definitions_path()}{os.sep}nxdlTypes.xsd"
    tree = ET.parse(filepath)
    root = tree.getroot()
    units_and_type_list = []
    for child in root:
        for i in child.attrib.values():
            units_and_type_list.append(i)
    flag = False
    for line in units_and_type_list:
        if line == 'anyUnitsAttr':
            flag = True
            nx_units = []
        elif 'NX' in line and flag is True:
            nx_units.append(line)
        elif line == 'primitiveType':
            flag = False
        else:
            pass
    return nx_units


def get_nx_attribute_type():
    """Read attribute types from the Nexus definition/nxdlTypes.xsd file

"""
    filepath = get_nexus_definitions_path() + '/nxdlTypes.xsd'
    tree = ET.parse(filepath)
    root = tree.getroot()
    units_and_type_list = []
    for child in root:
        for i in child.attrib.values():
            units_and_type_list.append(i)
    flag = False
    for line in units_and_type_list:
        if line == 'primitiveType':
            flag = True
            nx_types = []
        elif 'NX' in line and flag is True:
            nx_types.append(line)
        elif line == 'anyUnitsAttr':
            flag = False
        else:
            pass
    return nx_types


def get_node_name(node):
    '''Node - xml node
returns:
    html documentation name
    Either as specified by the 'name' or taken from the type (nx_class).
    Note that if only class name is available, the NX prefix is removed and
    the string is coverted to UPPER case.

'''
    if 'name' in node.attrib.keys():
        name = node.attrib['name']
    else:
        name = node.attrib['type']
        if name.startswith('NX'):
            name = name[2:].upper()
    return name


def belongs_to(nxdl_elem, child, hdf_name):
    """Checks if an HDF5 node name corresponds to a child of the NXDL element
uppercase letters in front can be replaced by arbitraty name, but
uppercase to lowercase match is preferred

"""
    # check if nameType allows different name
    try:
        name_any = bool(child.attrib['nameType'] == "any")
        # if child.attrib['nameType'] == "any":
        #    name_any = True
        # else:
        #    name_any = False
    except KeyError:
        name_any = False
    childname = get_node_name(child)
    nx_class_regex = re.compile(r"NX[a-z_]+")
    name = hdf_name[2:].upper() if nx_class_regex.search(hdf_name) and 'name\
    ' not in child.attrib else hdf_name
    # and no reserved words used
    if name_any and name != 'doc' and name != 'enumeration':
        # check if name fits
        fit = get_nx_namefit(name, childname)
        if fit < 0:
            return False
        for child2 in nxdl_elem:
            if get_local_name_from_xml(child) != \
                    get_local_name_from_xml(child2) or get_node_name(child2) == childname:
                continue
            # check if the name of another sibling fits better
            fit2 = get_nx_namefit(name, get_node_name(child2))
            if fit2 > fit:
                return False
        # accept this fit
        return True
    if childname == name:
        return True
    return False


def get_local_name_from_xml(element):
    """Helper function to extract the element tag without the namespace."""
    return element.tag[element.tag.rindex("}") + 1:]


def get_own_nxdl_child(nxdl_elem, name):
    """Checks if an NXDL child node fits to the specific name"""
    for child in nxdl_elem:
        if get_local_name_from_xml(child) == 'group' and belongs_to(nxdl_elem, child, name):
            # get_nx_class(child) == name:
            return child
        if get_local_name_from_xml(child) == 'field' and belongs_to(nxdl_elem, child, name):
            return child
        if get_local_name_from_xml(child) == 'attribute' and belongs_to(nxdl_elem, child, name):
            return child
        if get_local_name_from_xml(child) == 'doc' and name == 'doc':
            return child
        if get_local_name_from_xml(child) == 'enumeration' and name == 'enumeration':
            return child
    return None


def get_nxdl_child(nxdl_elem, name):
    """Get the NXDL child node corresponding to a specific name
(e.g. of an HDF5 node,or of a documentation)
note that if child is not found in application definition,
it also checks for the base classes

"""
    own_child = get_own_nxdl_child(nxdl_elem, name)
    if own_child is not None:
        return own_child
    # check in the base class
    bc_name = get_nx_class(nxdl_elem)
    # filter primitive types
    if bc_name[2] == '_':
        return None

    # Checking if it is actually the root element. Then we send it to NXroot.nxdl.xml
    if bc_name == "group":
        bc_name = "NXroot"
    bc_obj = ET.parse(f"{get_nexus_definitions_path()}{os.sep}"
                      f"base_classes{os.sep}{bc_name}.nxdl.xml").getroot()
    return get_own_nxdl_child(bc_obj, name)


def get_required_string(nxdl_elem):
    """Check for being required
REQUIRED, RECOMMENDED, OPTIONAL, NOT IN SCHEMA

"""
    if nxdl_elem is None:
        return "<<NOT IN SCHEMA>>"
    is_optional = 'optional' in nxdl_elem.attrib.keys() \
        and nxdl_elem.attrib['optional'] == "true"
    is_minoccurs = 'minOccurs' in nxdl_elem.attrib.keys() \
        and nxdl_elem.attrib['minOccurs'] == "0"
    # is_required = 'required' in nxdl_elem.attrib.keys() and nxdl_elem.attrib['required'] == "true"
    is_recommended = 'recommended' in nxdl_elem.attrib.keys() \
        and nxdl_elem.attrib['recommended'] == "true"

    if is_recommended:
        return "<<RECOMMENDED>>"
    if is_optional or is_minoccurs:
        return "<<OPTIONAL>>"
    return "<<REQUIRED>>"

    # default optionality
    # in BASE CLASSES: true
    # in APPLICATIONS: false

    # TODO: confirm with Sandor if we can get rid of these lines
    # if "base" in nxdl_elem.attrib and "base_classes" in nxdl_elem.base:
    #    return "<<OPTIONAL>>"


def chk_nxdataaxis_v2(hdf_node, name):
    """Check if dataset is an axis

"""
    # check for being a Signal
    own_signal = hdf_node.attrs.get('signal')
    if own_signal is str and own_signal == "1":
        LOGGER.info("Dataset referenced (v2) as NXdata SIGNAL")
    # check for being an axis
    own_axes = hdf_node.attrs.get('axes')
    if own_axes is str:
        axes = own_axes.split(':')
        for i in len(axes):
            if axes[i] and name == axes[i]:
                LOGGER.info("Dataset referenced (v2) as NXdata AXIS #%d", i)
                return None
    ownpaxis = hdf_node.attrs.get('primary')
    own_axis = hdf_node.attrs.get('axis')
    assert own_axis is int, 'axis is not int type!'
    # also convention v1
    if ownpaxis is int and ownpaxis == 1:
        LOGGER.info("Dataset referenced (v2) as NXdata AXIS #%d", own_axis - 1)
        return None
    if not (ownpaxis is int and ownpaxis == 1):
        LOGGER.info(
            "Dataset referenced (v2) as NXdata (primary/alternative) AXIS #%d", own_axis - 1)
        return None
    return None


def chk_nxdataaxis(hdf_node, name, loger):
    """NEXUS Data Plotting Standard v3: new version from 2014

"""
    # check if it is a field in an NXdata node
    if not isinstance(hdf_node, h5py.Dataset):
        return None
    parent = hdf_node.parent
    if not parent or (parent and not parent.attrs.get('NX_class') == "NXdata"):
        return None
    # chk for Signal
    signal = parent.attrs.get('signal')
    if signal and name == signal:
        loger.info("Dataset referenced as NXdata SIGNAL")
        return None
    # check for default Axes
    axes = parent.attrs.get('axes')
    if axes is str:
        if name == axes:
            loger.info("Dataset referenced as NXdata AXIS")
            return None
    elif axes is not None:
        for i, j in enumerate(axes):
            if name == j:
                indices = parent.attrs.get(j + '_indices')
                if indices is int:
                    loger.info("Dataset referenced as NXdata AXIS #%d" % indices)
                else:
                    loger.info("Dataset referenced as NXdata AXIS #%d" % i)
                return None
    # check for alternative Axes
    indices = parent.attrs.get(name + '_indices')
    if indices is int:
        loger.info("Dataset referenced as NXdata alternative AXIS #%d" % indices)
    # check for older conventions
    return chk_nxdataaxis_v2(hdf_node, name)


def get_nxdl_doc(hdf_node, loger, doc, attr=False):
    """Get nxdl documentation for an HDF5 node (or its attribute)
"""
    nxdl_file_path = (f"{get_nexus_definitions_path()}{os.sep}applications"
                      f"{os.sep}{get_nxdl_entry(hdf_node)}.nxdl.xml")
    elem = ET.parse(nxdl_file_path).getroot()
    nxdl_path = [elem]
    path = get_nx_class_path(hdf_node)
    req_str = None
    for group in path.split('/')[1:]:
        if group.startswith('NX'):
            elem = get_nxdl_child(elem, group)
            if elem is not None:
                if doc:
                    loger.info("/" + group)
                nxdl_path.append(elem)
            else:
                if doc:
                    loger.info("/" + group + " - IS NOT IN SCHEMA")
        else:
            if elem is not None:
                elem = get_nxdl_child(elem, group)
                nxdl_path.append(elem)
            if elem is not None:
                if attr:
                    if doc:
                        loger.info("/" + group)
                else:
                    if doc:
                        loger.info("/" + group + ' [' + get_nx_class(elem) + ']')
            else:
                if doc:
                    loger.info("/" + group + " - IS NOT IN SCHEMA")
    if elem is not None and attr:
        # NX_class is a compulsory attribute for groups in a nexus file
        # which should match the type of the corresponding NXDL element
        if attr == 'NX_class' and not isinstance(hdf_node, h5py.Dataset):
            elem = None
            if doc:
                loger.info("@" + attr + ' [NX_CHAR]')
        # units category is a compulsory attribute for any fields
        elif attr == 'units' and isinstance(hdf_node, h5py.Dataset):
            req_str = "<<REQUIRED>>"
            try:
                # try to find if units is defined inside the field in the NXDL element
                unit = elem.attrib[attr]
                if doc:
                    loger.info("@" + attr + ' [' + unit + ']')
                elem = None
                nxdl_path.append(attr)
            except KeyError:
                # otherwise try to find if units is defined as a child of the NXDL element
                elem = get_nxdl_child(elem, attr)
                if elem is not None:
                    if doc:
                        loger.info("@" + attr + ' - [' + get_nx_class(elem) + ']')
                    nxdl_path.append(elem)
                else:
                    # if no units category were defined in NXDL:
                    if doc:
                        loger.info("@" + attr + " - REQUIRED, but undefined unit category")
                    nxdl_path.append(attr)
                    # pass
        # units for attributes can be given as ATTRIBUTENAME_units
        elif attr.endswith('_units'):
            # check for ATTRIBUTENAME_units in NXDL (normal)
            elem2 = get_nxdl_child(elem, attr)
            if elem2 is not None:
                elem = elem2
                if doc:
                    loger.info("@" + attr + ' - [' + get_nx_class(elem) + ']')
                nxdl_path.append(elem)
            else:
                # if not defined, check for ATTRIBUTENAME to see if the ATTRIBUTE
                # is in the SCHEMA, but no units category were defined
                elem2 = get_nxdl_child(elem, attr[:-6])
                if elem2 is not None:
                    req_str = '<<RECOMMENDED>>'
                    if doc:
                        loger.info("@" + attr + " - RECOMMENDED, but undefined unit category")
                    nxdl_path.append(attr)
                else:
                    # otherwise: NOT IN SCHEMA
                    elem = elem2
                    if doc:
                        loger.info("@" + attr + " - IS NOT IN SCHEMA")
        # other attributes
        else:
            elem = get_nxdl_child(elem, attr)
            if elem is not None:
                if doc:
                    loger.info("@" + attr + ' - [' + get_nx_class(elem) + ']')
                nxdl_path.append(elem)
            else:
                if doc:
                    loger.info("@" + attr + " - IS NOT IN SCHEMA")

    if (elem is None and req_str is None):
        if doc:
            loger.info("")
        return ('None', None, None)
    if req_str is None:
        # check for being required
        req_str = get_required_string(elem)
        if doc:
            loger.info(req_str)
    if elem is not None:
        # check for deprecation
        dep_str = elem.attrib.get('deprecated')
        if dep_str:
            if doc:
                loger.info("DEPRECATED - " + dep_str)
        # check for enums
        sdoc = get_nxdl_child(elem, 'enumeration')
        if sdoc is not None:
            if doc:
                loger.info("enumeration:")
            for item in sdoc:
                if get_local_name_from_xml(item) == 'item':
                    if doc:
                        loger.info("-> " + item.attrib['value'])
        # check for NXdata references (axes/signal)
        chk_nxdataaxis(hdf_node, path.split('/')[-1], loger)
        # check for doc
        sdoc = get_nxdl_child(elem, 'doc')
        if doc:
            loger.info(get_local_name_from_xml(sdoc) if sdoc is not None else "")
    return (req_str, get_nxdl_entry(hdf_node), nxdl_path)


def get_doc(node, ntype, nxhtml, nxpath):
    """Get documentation

    """
    # URL for html documentation
    anchor = ''
    for n_item in nxpath:
        anchor += n_item.lower() + "-"
    anchor = ('https://manual.nexusformat.org/classes/',
              nxhtml + "#" + anchor.replace('_', '-') + ntype)
    if not ntype:
        anchor = anchor[:-1]

    # RST documentation from the field 'doc'
    doc = ""
    doc_field = node.find("doc")
    if doc_field is not None:
        doc = doc_field.text

    # enums
    (index, enums) = get_enums(node)
    if index:
        enum_str = "\n " + ("Possible values:"
                            if len(enums.split(',')) > 1
                            else "Obligatory value:") + "\n   " + enums + "\n"
    else:
        enum_str = ""

    return anchor, doc + enum_str


def print_doc(node, ntype, level, nxhtml, nxpath):
    """Print documentation
"""
    anchor, doc = get_doc(node, ntype, nxhtml, nxpath)
    print("  " * (level + 1) + anchor)

    preferred_width = 80 + level * 2
    wrapper = textwrap.TextWrapper(initial_indent='  ' * (level + 1), width=preferred_width,
                                   subsequent_indent='  ' * (level + 1), expand_tabs=False,
                                   tabsize=0)
    # doc=node.find('doc')
    if doc is not None:
        # for par in doc.text.split('\n'):
        for par in doc.split('\n'):
            print(wrapper.fill(par))
    # print(doc.text if doc is not None else "")


def get_namespace(element) -> str:
    """Extracts the namespace for elements in the NXDL"""
    return element.tag[element.tag.index("{"):element.tag.rindex("}") + 1]


def get_enums(node):
    """
    makes list of enumerations, if node contains any.
    Returns comma separated STRING of enumeration values, if there are enum tag,
    otherwise empty string.
    """
    # collect item values from enumeration tag, if any
    namespace = get_namespace(node)
    enums = []
    for enumeration in node.findall(f"{namespace}enumeration"):
        for item in enumeration.findall(f"{namespace}item"):
            enums.append(item.attrib["value"])
        enums = ','.join(enums)
        if enums != "":
            return (True, '[' + enums + ']')
    # if there is no enumeration tag, returns empty string
    return (False, "")


def get_node_at_nxdl_path(nxdl_path: str = None,
                          nx_name: str = None,
                          elem: ET.Element = None):
    """Returns an ET.Element for the given path.

This function either takes the name for the Nexus Application Definition
we are looking for or the root elem from a previously loaded NXDL file
and finds the corresponding XML element with the needed attributes.

"""
    if elem is None:
        nxdl_file_path = (f"{get_nexus_definitions_path()}{os.sep}"
                          f"applications{os.sep}{nx_name}.nxdl.xml")
        elem = ET.parse(nxdl_file_path).getroot()
    for group in nxdl_path.split('/')[1:]:
        elem = get_nxdl_child(elem, group)
    if elem is None:
        raise NxdlAttributeError(f"Attributes were not found for {nxdl_path}. "
                                 "Please check this entry in the template dictionary.")
    return elem


def process_node(hdf_node, hdf_path, parser, logger, doc=True):
    """
            #processes an hdf5 node
            #- it logs the node found and also checks for its attributes
            #- retrieves the corresponding nxdl documentation
            #TODO:
            # - follow variants
            # - NOMAD parser: store in NOMAD
            """
    hdf_info = {'hdf_path': hdf_path, 'hdf_node': hdf_node}
    if isinstance(hdf_node, h5py.Dataset):
        logger.info('===== FIELD (/%s): %s' % (hdf_path, hdf_node))
        val = str(hdf_node[()]).split('\n') if len(hdf_node.shape) <= 1 else str(
            hdf_node[0]).split('\n')
        logger.info('value: %s %s' % (val[0], "..." if len(val) > 1 else ''))
    else:
        logger.info('===== GROUP (/%s [%s::%s]): %s' %
                    (hdf_path, get_nxdl_entry(hdf_node),
                     get_nx_class_path(hdf_node), hdf_node))
    (req_str, nxdef, nxdl_path) = get_nxdl_doc(hdf_node, logger, doc)
    if parser is not None and isinstance(hdf_node, h5py.Dataset):
        parser(hdf_info, nxdef, nxdl_path, val)
    for key, value in hdf_node.attrs.items():
        logger.info('===== ATTRS (/%s@%s)' % (hdf_path, key))
        val = str(value).split('\n')
        logger.info('value: %s %s' % (val[0], "..." if len(val) > 1 else ''))
        (req_str, nxdef, nxdl_path) = get_nxdl_doc(hdf_node, logger, doc, attr=key)
        if parser is not None and 'NOT IN SCHEMA' not in req_str and 'None' not in req_str:
            parser(hdf_info, nxdef, nxdl_path, val)


def get_default_plotable(root, logger):
    """Get default plotable

"""
    logger.info('========================')
    logger.info('=== Default Plotable ===')
    logger.info('========================')
    # v3 from 2014

    # nxentry
    nxentry = None
    default_nxentry_group_name = root.attrs.get("default")
    if default_nxentry_group_name:
        try:
            nxentry = root[default_nxentry_group_name]
        except KeyError:
            nxentry = None
    if not nxentry:
        nxentry = entry_helper(root)
    if not nxentry:
        logger.info('No NXentry has been found')  # TODO the code always falls here, solve it
        return
    logger.info('')
    logger.info('NXentry has been identified: ' + nxentry.name)
    # process_node(nxentry, None, False)
    # nxdata
    nxdata = None
    default_nxdata_group_name = nxentry.attrs.get("default")
    if default_nxdata_group_name:
        try:
            nxdata = nxentry[default_nxdata_group_name]
        except KeyError:
            nxdata = None
    if not nxdata:
        nxdata = nxdata_helper(nxentry)
    if not nxdata:
        logger.info('No NXdata group has been found')
        return
    logger.info('')
    logger.info('NXdata group has been identified: ' + nxdata.name)
    process_node(nxdata, nxdata.name, None, logger, False)
    # signal
    signal = None
    signal_dataset_name = nxdata.attrs.get("signal")
    try:
        signal = nxdata[signal_dataset_name]
    except KeyError:
        signal = None
    if not signal:
        signal = signal_helper(nxdata)
    if not signal:
        logger.info('No Signal has been found')
        return
    logger.info('')
    logger.info('Signal has been identified: ' + signal.name)
    process_node(signal, signal.name, None, logger, False)
    dim = len(signal.shape)
    # axes
    axes = []
    axis_helper(dim, nxdata, signal, axes, logger)


def entry_helper(root):
    """Check entry related data

"""
    nxentries = []
    for key in root.keys():
        if (isinstance(root[key], h5py.Group) and root[key].attrs.get('NX_class\
        ') and root[key].attrs['NX_class'] == "NXentry"):
            nxentries.append(root[key])
    # v3: as there was no selection given, only 1 nxentry shall exists
    # v2: take any
    if len(nxentries) >= 1:
        return nxentries[0]
    return None


def nxdata_helper(nxentry):
    """Check if nxentry hdf5 object has a NX_class and, if it contains NXdata,
return its value

"""
    lnxdata = []
    for key in nxentry.keys():
        if isinstance(nxentry[key], h5py.Group) and nxentry[key].attrs.get('NX_class\
        ') and nxentry[key].attrs['NX_class'] == "NXdata":
            lnxdata.append(nxentry[key])
    # v3: as there was no selection given, only 1 nxdata shall exists
    # v2: take any
    if len(lnxdata) >= 1:
        return lnxdata[0]
    return None


def signal_helper(nxdata):
    """Check signal related data

"""
    signals = []
    for key in nxdata.keys():
        if isinstance(nxdata[key], h5py.Dataset):
            signals.append(nxdata[key])
    # v3: as there was no selection given, only 1 data field shall exists
    if len(signals) == 1:
        return signals[0]
    # v2: select the one with an attribute signal="1" attribute
    if len(signals) > 1:
        for sig in signals:
            if sig.attrs.get("signal") and sig.attrs.get("signal\
            ") is str and sig.attrs.get("signal") == "1":
                return sig
    return None


def find_attrib_axis_actual_dim_num(nxdata, a_item, ax_list):
    """Finds axis that have defined dimensions"""
    # find those with attribute axis= actual dimension number
    lax = []
    for key in nxdata.keys():
        if isinstance(nxdata[key], h5py.Dataset):
            try:
                if nxdata[key].attrs['axis'] == a_item + 1:
                    lax.append(nxdata[key])
            except KeyError:
                pass
    if len(lax) == 1:
        ax_list.append(lax[0])
    # if there are more alternatives, prioritise the one with an attribute primary="1"
    elif len(lax) > 1:
        for sax in lax:
            if sax.attrs.get('primary') and sax.attrs.get('primary') == 1:
                ax_list.insert(0, sax)
            else:
                ax_list.append(sax)


def get_single_or_multiple_axes(nxdata, ax_datasets, a_item, ax_list):
    """Gets either single or multiple axes from the NXDL"""
    try:
        # single axis is defined
        if ax_datasets is str:
            # explicite definition of dimension number
            ind = nxdata.attrs.get(ax_datasets + '_indices')
            if ind and ind is int:
                if ind == a_item:
                    ax_list.append(nxdata[nxdata[ax_datasets]])
            # positional determination of the dimension number
            elif a_item == 0:
                ax_list.append(nxdata[nxdata[ax_datasets]])
        # multiple axes are listed
        else:
            # explicite definition of dimension number
            for aax in ax_datasets:
                ind = nxdata.attrs.get(aax + '_indices')
                if ind and ind is int:
                    if ind == a_item:
                        ax_list.append(nxdata[nxdata[aax]])
            # positional determination of the dimension number
            if not ax_list:
                ax_list.append(nxdata[ax_datasets[a_item]])
    except KeyError:
        pass


def axis_helper(dim, nxdata, signal, axes, logger):
    """Check axis related data

"""
    for a_item in range(dim):
        ax_list = []
        # primary axes listed in attribute axes
        ax_datasets = nxdata.attrs.get("axes")
        get_single_or_multiple_axes(nxdata, ax_datasets, a_item, ax_list)
        # check for corresponding AXISNAME_indices
        for attr in nxdata.attrs.keys():
            if attr.endswith('_indices') and nxdata.sttrs[attr] == a_item:
                ax_list.append(nxdata[attr.split('_indices')[0]])
        # v2
        # check for ':' separated axes defined in Signal
        if not ax_list:
            try:
                ax_datasets = signal.attrs.get("axes").split(':')
                ax_list.append(nxdata[ax_datasets[a_item]])
            except KeyError:
                pass
        # check for axis/primary specifications
        if not ax_list:
            find_attrib_axis_actual_dim_num(nxdata, a_item, ax_list)
        axes.append(ax_list)
        logger.info('')
        logger.info('For Axis #%d, %d axes have been identified: %s\
        ' % (a_item, len(ax_list), str(ax_list)))


class HandleNexus:
    """documentation

"""
    def __init__(self, logger, args):
        self.logger = logger
        self.input_file_name = args[0] if len(
            # args) >= 1 else 'wcopy/from_dallanto_master_from_defs.h5'
            # args) >= 1 else 'ARPES/201805_WSe2_arpes.nxs'
            args) >= 1 else 'tests/data/nexus_test_data/201805_WSe2_arpes.nxs'
        self.parser = None
        self.in_file = None

    def visit_node(self, hdf_name, hdf_node):
        """Function called by h5py that iterates on each node of hdf5file.

        It allows h5py visititems function to visit nodes.

        """
        hdf_path = '/' + hdf_name
        process_node(hdf_node, hdf_path, self.parser, self.logger)

    def process_nexus_master_file(self, parser):
        """
        Process a nexus master file by processing all its nodes and their attributes

        """
        self.parser = parser
        self.in_file = h5py.File(self.input_file_name, 'r')
        self.in_file.visititems(self.visit_node)
        get_default_plotable(self.in_file, self.logger)
        self.in_file.close()


if __name__ == '__main__':
    LOGGING_FORMAT = "%(levelname)s: %(message)s"
    STDOUT_HANDLER = logging.StreamHandler(sys.stdout)
    STDOUT_HANDLER.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.DEBUG, format=LOGGING_FORMAT, handlers=[STDOUT_HANDLER])
    LOGGER = logging.getLogger()
    NEXUS_HELPER = HandleNexus(LOGGER, sys.argv[1:])
    NEXUS_HELPER.process_nexus_master_file(None)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import logging
import numpy as np
from datetime import datetime

from nomad.units import ureg
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.simulation.run import (
    Run, Program, TimeRun
)
from nomad.datamodel.metainfo.simulation.method import (
    Method, DFT, XCFunctional, Functional, BasisSet
)
from nomad.datamodel.metainfo.simulation.system import (
    System, Atoms
)
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, Energy, EnergyEntry, ScfIteration, Forces, ForcesEntry, Stress,
    StressEntry, Multipoles, MultipolesEntry, Charges, ChargesValue
)
from nomad.datamodel.metainfo.workflow import Workflow


re_n = r'[\n\r]'
re_f = r'[-+]?\d+\.\d*(?:[DdEe][-+]\d+)?'


class FDFParser(TextParser):
    def init_quantities(self):
        def to_parameter(val_in):
            key, val = val_in.strip().split(' ', 1)
            if val[0].isdecimal() or val[0] == '-':
                val = val.split()[0]
            elif val.lower() in ['true', 'false']:
                val = val == 'true'
            return [key, val]

        def to_block(val_in):
            key, val = val_in.strip().split(' ', 1)
            return [key, [v.strip().split() for v in val.strip().splitlines()]]

        self._quantities = [
            Quantity(
                'parameter', rf'([A-Za-z][\w\.]+) +([ \w\.\/]+?)(?:\#|{re_n})',
                repeats=True, str_operation=to_parameter
            ),
            Quantity(
                'block', rf'\%block ([A-Za-z][\w\.]+)\s+([\s\S]+?)\%endblock',
                repeats=True, str_operation=to_block
            )
        ]

    def parse(self, key=None):
        super().parse(key)

        # add atoms info
        species, coordinates, lattice_vectors = None, None, None
        parameters = {key: val for key, val in self._results.get('parameter', [])}
        self._results['parameters'] = parameters

        # scaling of the lattice vectors
        lattice_constant = parameters.get('LatticeConstant', 1.0) * ureg.angstrom
        # units of the atomic coordinates
        coordinates_format = parameters.get('AtomicCoordinatesFormat')

        for block in self._results.get('block', []):
            if block[0] == 'ChemicalSpeciesLabel':
                species = [b[2] for b in block[1]]
            elif block[0] == 'LatticeVectors':
                lattice_vectors = block[1] * lattice_constant
            elif block[0] == 'AtomicCoordinatesAndAtomicSpecies':
                coordinates = np.array([b[0:3] for b in block[1]], dtype=np.float64)

        self._results['positions'] = coordinates
        self._results['labels'] = species
        self._results['lattice_vectors'] = lattice_vectors
        self._results['coordinates_format'] = coordinates_format


class OutParser(TextParser):
    def init_quantities(self):
        def to_stress_tensor(val_in):
            val = val_in.strip().split()
            stress = np.zeros((3, 3))
            stress[0][0] = val[0]
            stress[1][1] = val[1]
            stress[2][2] = val[2]
            stress[0][1] = stress[1][0] = val[3]
            stress[1][2] = stress[2][1] = val[4]
            stress[0][2] = stress[2][0] = val[5]
            return stress * ureg.kbar

        calc_quantities = [
            Quantity(
                'lattice_vectors',
                r'outcell\: Unit cell vectors \(Ang\)\:\s+([\s\d\.\-]+)',
                shape=(3, 3), dtype=np.dtype(np.float64), unit=ureg.angstrom
            ),
            Quantity(
                'atoms',
                r'outcoor: Atomic coordinates(.+[\s\w\.\-]+)',
                sub_parser=TextParser(quantities=[
                    Quantity('coordinates_format', r'\((\S+)\)', dtype=str),
                    Quantity('labels', r' +([A-Z][a-z]*)', repeats=True, dtype=str),
                    Quantity('positions', rf'({re_f} +{re_f} +{re_f})', repeats=True, dtype=np.dtype(np.float64))
                ])
            ),
            Quantity(
                'energy',
                rf'Program\'s energy decomposition \(eV\)\:([\s\S]+?){re_n} *{re_n}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'contribution', rf'siesta\: +(\S+) *= *({re_f})',
                        repeats=True
                    )
                ])
            ),
            Quantity(
                'scf',
                rf'scf\: iscf +Eharris\(eV\).+([\s\S]+?){re_n} *{re_n}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'step', rf'scf\: +\d+ +({re_f} +{re_f} +{re_f} +{re_f} +{re_f})',
                        repeats=True, dtype=np.dtype(np.float64))
                ])
            ),
            Quantity(
                'energy_total', rf'siesta\: E_KS\(eV\) = +({re_f})',
                dtype=np.float64, unit=ureg.eV
            ),
            Quantity(
                'energy',
                rf'Final energy \(eV\)\:([\s\S]+?Total.+)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'contribution', rf'siesta: +([ \w\.\-]+?) *= *({re_f})',
                        repeats=True, str_operation=lambda x: [v.strip() for v in x.rsplit(' ', 1)]
                    )
                ])
            ),
            Quantity(
                'forces',
                r'siesta\: Atomic forces \(eV/Ang\)\:([\s\S]+?Tot.+)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'atomic', rf'\d+ +({re_f} +{re_f} +{re_f})',
                        repeats=True, dtype=np.dtype(np.float64)
                    )
                ])
            ),
            Quantity(
                'stress_tensor',
                rf'Stress\-tensor\-Voigt \(kbar\)\: +({re_f} +{re_f} +{re_f} +{re_f} +{re_f} +{re_f})',
                str_operation=to_stress_tensor
            ),
            Quantity(
                'stress_tensor',
                rf'Stress tensor \(static\) \(eV/Ang\*\*3\)\:([\s\S]+?){re_n} *{re_n}',
                str_operation=lambda x: [v.strip().split()[1:4] for v in x.strip().splitlines()],
                dtype=np.dtype(np.float64), unit=ureg.eV / ureg.angstrom ** 3
            ),
            Quantity(
                'electric_dipole',
                rf'Electric dipole \(Debye\) += +({re_f} +{re_f} +{re_f})',
                dtype=np.dtype(np.float64)
            ),
            Quantity(
                'mulliken',
                r'mulliken\: Atomic and Orbital Populations\:([\s\S]+?)siesta\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'atom',
                        rf'(Species\:[\s\S]+?){re_n} *{re_n}',
                        repeats=True, sub_parser=TextParser(quantities=[
                            Quantity('label', r'Species\: *(\S+)', dtype=str),
                            Quantity('orbital', r'(\d[spdSPD]\S*) ', dtype=str, repeats=True),
                            Quantity('values', rf'(\d+ +{re_f} +{re_f}[\d\.\-\s]+)', dtype=np.dtype(np.float64))
                        ])
                    ),
                    Quantity('spin', r'(Spin DOWN)', str_operation=lambda x: 2),
                    Quantity('q_tot', rf'Qtot = +({re_f})', repeats=True, dtype=np.float64)
                ])
            )
        ]

        self._quantities = [
            Quantity(
                'header',
                r'(S[\s\S]+?)\*',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'program_version', r'(?:SIESTA|Siesta Version:) *(.+)',
                        flatten=False, dtype=str
                    ),
                    Quantity(
                        'x_siesta_arch', r'Architecture *: +(.+)',
                        flatten=False, dtype=str
                    ),
                    Quantity(
                        'x_siesta_compiler_flags', r'Compiler flags *: +(.+)',
                        flatten=False, dtype=str
                    ),
                    Quantity(
                        'x_siesta_parallel', r'(PARALLEL) version',
                        str_operation=lambda x: True
                    ),
                ])
            ),
            Quantity(
                'input_data_file',
                r'(Dump of input data file.+[\s\S]+?End of input data file)',
                flatten=False, dtype=str
            ),
            Quantity(
                'simulation_parameters',
                r'Simulation parameters.+([\s\S]+?)\*{10}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'redata', r'redata\: (.+?= +\S+)',
                        str_operation=lambda x: [v.strip() for v in x.rsplit('=')], repeats=True
                    )
                ])
            ),
            Quantity(
                'x_siesta_n_nodes', r'Running on +(\d+) nodes',
                dtype=np.int32
            ),
            Quantity(
                'run_start', r'Start of run: +(\d+-[A-Z]+-\d+) +(\d+:\d+:\d+)',
                flatten=False, dtype=str
            ),
            Quantity(
                'run_end', r'End of run: +(\d+-[A-Z]+-\d+) +(\d+:\d+:\d+)',
                flatten=False, dtype=str
            ),
            Quantity(
                'single_point',
                r'(Single\-point calculation[\s\S]+?Target enthalpy.+)',
                sub_parser=TextParser(quantities=calc_quantities)
            ),
            Quantity(
                'geometry_optimization',
                r'\={5}\s+(Begin[\s\S]+?)outcoor\: Re',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'step',
                        r'(Begin.+?opt\. move += +\d+[\s\S]+?Target enthalpy.+)',
                        repeats=True, sub_parser=TextParser(quantities=calc_quantities)
                    )
                ])
            ),
        ] + calc_quantities


class SiestaParser:
    def __init__(self):
        self.out_parser = OutParser()
        self.fdf_parser = FDFParser()
        self._xc_map = {
            'ca': ('LDA_X', 'LDA_C_PZ'),
            'pz': ('LDA_X', 'LDA_C_PZ'),
            'pw92': ('LDA_X', 'LDA_C_PW'),
            # TODO 'PW91': '',
            'pbe': ('GGA_X_PBE', 'GGA_C_PBE'),
            'revpbe': ('GGA_X_PBE_R', 'GGA_C_PBE'),
            'rpbe': ('GGA_X_RPBE', 'GGA_C_PBE'),
            # TODO 'WC': ('GGA_X_WC', ),
            'am05': ('GGA_X_AM05', 'GGA_C_AM05'),
            'pbesol': ('GGA_X_PBE_SOL', 'GGA_C_PBE_SOL'),
            'blyp': ('GGA_X_B88', 'GGA_C_LYP'),
            'df1': ('gga_x_pbe_r', 'vdw_c_df1'),
            'drsll': ('gga_x_pbe_r', 'vdw_c_df1'),
            'lmkll': ('gga_x_rpw86', 'vdw_c_df2'),
            'df2': ('gga_x_rpw86', 'vdw_c_df2'),
            'kbm': ('GGA_X_OPTB88_VDW', 'vdw_c_df1'),
            'c09': ('GGA_X_C09X', 'vdw_c_df1'),
            'bh': ('GGA_X_LV_RPW86', 'vdw_c_df1'),
        }
        self._metainfo_map = {
            'Etot': 'total', 'FreeEng': 'free', 'Exc': 'xc', 'Band Struct.': 'sum_eigenvalues',
            'Hartree': 'electrostatic', 'Exch.-corr.': 'xc', 'Ion-ion': 'nuclear_repulsion',
            'Total': 'total'
        }

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger
        self.fdf_parser.logger = self.logger
        self._files = os.listdir(self.maindir)

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logging.getLogger(__name__) if logger is None else logger
        self.maindir = os.path.dirname(self.filepath)
        self.init_parser()

        sec_run = archive.m_create(Run)
        header = self.out_parser.get('header')
        if header is not None:
            sec_run.program = Program(name='Siesta', version=header.program_version)
            for key, val in header.items():
                if key.startswith('x_siesta'):
                    setattr(sec_run, key, val)

        for key, val in self.out_parser.items():
            if key.startswith('x_siesta'):
                setattr(sec_run, key, val)

        def get_date(time):
            if time is None:
                return
            return datetime.strptime(time, '%d-%b-%Y %H:%M:%S').timestamp()

        sec_run.time_run = TimeRun(
            date_start=get_date(self.out_parser.run_start), date_end=get_date(self.out_parser.run_end))

        # input parameters from data file
        self.fdf_parser.mainfile = self.out_parser.mainfile
        if self.out_parser.input_data_file is not None:
            self.fdf_parser._file_handler = self.out_parser.input_data_file.encode()

        sec_method = sec_run.m_create(Method)
        parameters = self.fdf_parser.get('parameters')
        sec_method.x_siesta_input_parameters = parameters
        sec_method.x_siesta_simulation_parameters = {
            key: val for key, val in self.out_parser.get('simulation_parameters', {}).get('redata', [])}
        sec_method.dft = DFT(xc_functional=XCFunctional())
        for xc_functional in self._xc_map.get(parameters.get('xc.authors').lower(), []):
            if '_X_' in xc_functional:
                sec_method.dft.xc_functional.exchange.append(Functional(name=xc_functional))
            elif '_C_' in xc_functional:
                sec_method.dft.xc_functional.correlation.append(Functional(name=xc_functional))
            elif '_HYB_' in xc_functional:
                sec_method.dft.xc_functional.hybrid.append(Functional(name=xc_functional))
            else:
                sec_method.dft.xc_functional.contributions.append(Functional(name=xc_functional))
        sec_basis_set = sec_method.m_create(BasisSet)
        sec_basis_set.type = 'numeric AOs'
        # parse atomic basis info

        def parse_system(source):
            sec_system = sec_run.m_create(System)
            lattice_vectors = source.get('lattice_vectors')
            atoms = source.get('atoms')
            source = atoms if atoms is not None else source
            positions = source.get('positions')
            if positions is not None:
                coordinates_format = source.get('coordinates_format', 'Ang').lower()
                if coordinates_format == 'ang':
                    positions = positions * ureg.angstrom
                elif coordinates_format in ['fractional', 'scaledcartesian']:
                    if lattice_vectors is not None:
                        positions = np.dot(positions, lattice_vectors)

            sec_system.atoms = Atoms(
                positions=positions, labels=source.get('labels'),
                lattice_vectors=lattice_vectors)

            return sec_system

        def parse_calculation(source):
            sec_calc = sec_run.m_create(Calculation)
            # energy
            sec_energy = sec_calc.m_create(Energy)
            for key, val in source.get('energy', {}).get('contribution', []):
                if key in self._metainfo_map:
                    setattr(sec_energy, self._metainfo_map[key], EnergyEntry(value=val * ureg.eV))
                else:
                    sec_energy.contributions.append(EnergyEntry(kind=key, value=val * ureg.eV))
            if source.energy_total is not None:
                sec_energy.total = EnergyEntry(value=source.energy_total)

            # forces
            if source.forces is not None and source.forces.atomic is not None:
                sec_calc.forces = Forces(total=ForcesEntry(value=source.forces.atomic * ureg.eV / ureg.angstrom))

            # stress tensor
            if source.stress_tensor is not None:
                sec_calc.stress = Stress(total=StressEntry(value=source.stress_tensor))

            # dipoles
            if source.electric_dipole is not None:
                sec_multipoles = sec_calc.m_create(Multipoles)
                sec_multipoles.kind = 'electric'
                sec_multipoles.dipole = MultipolesEntry(total=source.electric_dipole)

            # multipoles
            if source.mulliken is not None:
                sec_charges = sec_calc.m_create(Charges)
                atoms = source.mulliken.get('atom', [])
                spin = source.mulliken.get('spin', 1)
                n_atoms = len(atoms) // spin
                value = np.zeros(n_atoms)
                for atom_n, atom in enumerate(atoms):
                    spin_n = atom_n // n_atoms
                    for orb_n, orb in enumerate(atom.get('orbital', [])):
                        # TODO map orb to lm value
                        sec_charges.orbital_projected.append(ChargesValue(
                            spin=spin_n, atom_label=atom.label, orbital=orb,
                            value=atom.values[orb_n + 2] * ureg.elementary_charge  # 0: index 1: sum
                        ))
                    if spin == 2:
                        # spin peojected
                        sec_charges.spin_projected.append(ChargesValue(
                            spin=spin_n, atom_label=atom.label, value=atom.values[1] * ureg.elementary_charge
                        ))
                    value[atom_n % n_atoms] += atom.values[1]

                sec_charges.analysis_method = 'Mulliken'
                sec_charges.value = value * ureg.elementary_charge
                sec_charges.total = sum(source.mulliken.get('qtot', [])) * ureg.elementary_charge

            # TODO parse more properties
            # scf steps
            for step in source.get('scf', {}).get('step', []):
                sec_scf = sec_calc.m_create(ScfIteration)
                sec_scf.energy = Energy(
                    total=EnergyEntry(value=step[1] * ureg.eV), fermi=step[4] * ureg.eV,
                    free=EnergyEntry(value=step[2] * ureg.eV))
                sec_scf.energy.types.append(EnergyEntry(kind='Harris', value=step[0] * ureg.eV))

            return sec_calc

        def parse_configurations(source):
            system = parse_system(source)
            calc = parse_calculation(source)
            calc.system_ref = system

        workflow = archive.m_create(Workflow)
        if self.out_parser.single_point is not None:
            parse_configurations(self.out_parser.single_point)
            workflow.type = 'single_point'

        if self.out_parser.geometry_optimization is not None:
            for step in self.out_parser.geometry_optimization.get('step', []):
                parse_configurations(step)
            workflow.type = 'geometry_optimization'

        # final properties
        calc = parse_calculation(self.out_parser)
        calc.system_ref = sec_run.system[-1]




# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import os
import click

from .cli import cli


@cli.group(help='Commands related to the nomad source code.')
def dev():
    pass


@dev.command(help='Runs tests and linting of the nomad python source code. Useful before committing code.')
@click.option('--skip-tests', help='Do no tests, just do code checks.', is_flag=True)
@click.option('-x', '--exitfirst', help='Stop testing after first failed test case.', is_flag=True)
def qa(skip_tests: bool, exitfirst: bool):
    os.chdir(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    ret_code = 0

    if not skip_tests:
        click.echo('Run tests ...')
        ret_code += os.system('python -m pytest -sv%s tests' % ('x' if exitfirst else ''))
    click.echo('Run code style checks ...')
    ret_code += os.system('python -m pycodestyle --ignore=E501,E701,E731 nomad tests')
    click.echo('Run linter ...')
    ret_code += os.system('python -m pylint --load-plugins=pylint_mongoengine,nomad.metainfo.pylint_plugin nomad tests')
    click.echo('Run static type checks ...')
    ret_code += os.system('python -m mypy --ignore-missing-imports --follow-imports=silent --no-strict-optional nomad tests')

    sys.exit(ret_code)


@dev.command(help='Runs tests and linting of the nomad gui source code. Useful before committing code.')
@click.option('--skip-tests', help='Do no tests, just do code checks.', is_flag=True)
def gui_qa(skip_tests: bool):
    os.chdir(os.path.abspath(os.path.join(os.path.dirname(__file__), '../../gui')))
    ret_code = 0

    if not skip_tests:
        click.echo('Run gui testing ...')
        ret_code += os.system('yarn run test')
    click.echo('Run gui code linting ...')
    ret_code += os.system('yarn run lint')

    sys.exit(ret_code)


@dev.command(help='Generates a JSON with all metainfo.')
def metainfo():
    import json
    export = _all_metainfo_packages()
    metainfo_data = export.m_to_dict(with_meta=True)
    print(json.dumps(metainfo_data, indent=2))


def _all_metainfo_packages():
    from nomad.metainfo import Package, Environment
    from nomad.datamodel import EntryArchive

    # TODO similar to before, due to lazyloading, we need to explicily access parsers
    # to actually import all parsers and indirectly all metainfo packages
    from nomad.parsing.parsers import import_all_parsers
    import_all_parsers()

    # Create the ES mapping to populate ES annoations with search keys.
    from nomad.search import entry_type
    entry_type.create_mapping(EntryArchive.m_def)

    # TODO this is otherwise not imported and will add nexus to the Package.registry
    from nexusparser.metainfo import nexus  # pylint: disable=unused-import

    # TODO we call __init_metainfo__() for all packages where this has been forgotten
    # by the package author. Ideally this would not be necessary and we fix the
    # actual package definitions.
    for module_key in list(sys.modules):
        pkg: Package = getattr(sys.modules[module_key], 'm_package', None)
        if pkg is not None and isinstance(pkg, Package):
            if (pkg.name not in Package.registry):
                pkg.__init_metainfo__()

    export = Environment()
    for package in Package.registry.values():
        export.m_add_sub_section(Environment.packages, package)
    return export


@dev.command(help='Generates a JSON with all search quantities.')
def search_quantities():
    _all_metainfo_packages()
    import json

    # Currently only quantities with "entry_type" are included.
    from nomad.metainfo.elasticsearch_extension import entry_type, Elasticsearch
    from nomad.datamodel import EntryArchive

    def to_dict(search_quantity, section=False):
        if section:
            keys = ["name", "description", "nested"]
            metadict = search_quantity.sub_section.m_to_dict(with_meta=True)
            metadict["name"] = search_quantity.m_to_dict(with_meta=True)["name"]
            es_annotations = search_quantity.m_get_annotations(Elasticsearch, as_list=True)
            nested = any([x.nested for x in es_annotations])
            metadict["nested"] = nested
        else:
            keys = ["name", "description", "type", "unit"]
            metadict = search_quantity.definition.m_to_dict(with_meta=True)
        result = {}
        for key in keys:
            val = metadict.get(key)
            if val is not None:
                result[key] = val
        return result

    export = {}

    # Add quantities
    for search_quantity in entry_type.quantities.values():
        isSuggestion = search_quantity.annotation.suggestion
        if not isSuggestion:
            export[search_quantity.qualified_name] = to_dict(search_quantity)

    # Add suggestion flag
    for suggestion in entry_type.suggestions.keys():
        export[suggestion]["suggestion"] = True

    # Add section definitions
    def get_sections(m_def, prefix=None):
        for sub_section_def in m_def.all_sub_sections.values():
            name = sub_section_def.name
            full_name = f"{prefix}.{name}" if prefix else name
            export[full_name] = to_dict(sub_section_def, True)
            get_sections(sub_section_def.sub_section, full_name)
    get_sections(EntryArchive.results.sub_section, "results")

    print(json.dumps(export, indent=2))


@dev.command(help='Generates a JSON file that compiles all the parser metadata from each parser project.')
def parser_metadata():
    import json
    import yaml
    import os
    import os.path
    from glob import glob

    parsers_metadata = {}
    parsers_path = './dependencies/parsers'
    for parser_metadata_file in sorted(glob(f'{parsers_path}/**/metadata.yaml', recursive=True)):
        with open(parser_metadata_file) as f:
            parser_metadata = yaml.load(f, Loader=yaml.FullLoader)
        parsers_metadata[os.path.basename(os.path.dirname(parser_metadata_file))] = parser_metadata

    parsers_metadata = {
        key: parsers_metadata[key]
        for _, key in sorted([(key.lower(), key) for key in parsers_metadata], key=lambda x: x[0])}

    print(json.dumps(parsers_metadata, indent=2))


@dev.command(help='Generate toolkit tutorial metadata from anaytics submodules.')
def toolkit_metadata():
    import requests
    import re
    import json
    modules = requests.get(
        'https://gitlab.mpcdf.mpg.de/api/v4/projects/3161/repository/files/.gitmodules/raw?ref=master').text

    tutorials = []
    lines = modules.split('\n')
    for line in lines:
        match = re.match(r'\s*url = (.*)$', line)
        if match:
            url = match.group(1).replace('.git', '') + '/-/raw/master/metainfo.json'
            response = requests.get(url)
            if response.status_code != 200:
                print('Could not get metadata for %s' % match.group(1), file=sys.stderr)
                continue
            try:
                tutorials.append(response.json())
            except Exception:
                print('Could not get metadata for %s. Project is probably not public.' % match.group(1), file=sys.stderr)

    print(json.dumps(dict(tutorials=tutorials), indent=2))


@dev.command(help=(
    'Updates parser`s README files by combining a general template with  '
    'a parser`s metadata YAML file.'))
@click.option('--parser', help='Only updated the README of the given parsers subdirctory.')
def update_parser_readmes(parser):
    from glob import glob
    import re
    import yaml

    os.chdir(os.path.join(os.path.dirname(__file__), '../..'))

    # filenames
    local_fn = 'README.md'
    generic_fn = './README.parsers.md'
    parser_path = './dependencies/parsers/'

    # Open general template
    with open(generic_fn, 'r') as generic:  # read only
        generic_contents = generic.read()

    # Replace the comment at the top of the gereral template
    generic_contents = re.sub(
        rf'\*\*\*Note:\*\* This is a general README file for NOMAD parsers, '
        rf'consult the README of specific parser projects for more detailed '
        rf'information!\*\n\n', '', generic_contents)

    def open_metadata(path):
        # read local yaml metadata file
        with open(path, 'r') as metadata_f:
            try:
                metadata = yaml.load(metadata_f, Loader=yaml.FullLoader)
            except Exception as e:
                print(f'Error reading metadata.yaml: {e}')
                metadata = None
        return metadata

    def replace(metadata, contents, path):
        # replace placelder in contents with metadata values
        for key in metadata.keys():
            replace = metadata.get(key)
            if replace is None:
                continue
            print(f'\tReplacing {key} with {replace}')
            contents = re.sub(rf'\${key}\$', replace, contents)

        # save file
        with open(path, 'w') as f:
            f.write(contents.strip())
            f.write('\n')

    for local_readme in sorted(glob(f'{parser_path}/*/{local_fn}')):
        parser_dir = os.path.dirname(local_readme)
        project_name = os.path.basename(parser_dir)
        print(f'Working on {parser_dir}')

        contents = generic_contents
        # if metadata file under the parser directory exists, it is a single parser project
        single = os.path.isfile(os.path.join(parser_dir, 'metadata.yaml'))

        if single:
            metadata = open_metadata(os.path.join(parser_dir, 'metadata.yaml'))
            # git path is given by nomad-parser-codename
            metadata['gitPath'] = f'nomad-parser-{project_name}'
            parser_header, parser_specs = '', ''
        else:
            # replace header for the single parser with that for a group of parsers
            parser_header_re = r'(\nThis is a NOMAD parser[\s\S]+?Archive format\.\n)'
            parser_header = re.search(parser_header_re, contents).group(1)
            group_header = 'This is a collection of the NOMAD parsers for the following $codeName$ codes:\n\n$parserList$'
            contents = re.sub(parser_header_re, group_header, contents)
            # remove individual parser specs
            parser_specs_re = r'(For \$codeLabel\$ please provide[\s\S]+?\$tableOfFiles\$)\n\n'
            parser_specs = re.search(parser_specs_re, contents).group(1)
            contents = re.sub(parser_specs_re, '', contents)
            metadata = dict(
                gitPath=f'{project_name}-parsers',
                parserGitUrl=f'https://github.com/nomad-coe/{project_name}-parsers.git',
                parserSpecific='')
        if metadata.get('codeName', '').strip() == '':
            metadata['codeName'] = project_name
        if 'preamble' not in metadata:
            metadata['preamble'] = ''

        # if this is a group of parser, find all individdual parsers and write the
        # parser specs
        parser_list = ''
        for index, local_metadata in enumerate(sorted(glob(f'{parser_dir}/*/*/metadata.yaml'))):
            metadata_parser = open_metadata(local_metadata)
            # contents is simply the parser header and specs
            contents_parser = f'{parser_header}\n{parser_specs}'
            replace(metadata_parser, contents_parser, os.path.join(os.path.dirname(local_metadata), local_fn))
            # add the codename to the list of parsers for the group header
            codelabel = metadata_parser.get('codeLabel', os.path.basename(os.path.dirname(local_metadata)))
            codeurl = metadata_parser.get('codeUrl', '')
            parser_list = rf'{parser_list}{index + 1}. [{codelabel}]({codeurl})\n'
        metadata['parserList'] = parser_list.strip()

        # Find & Replace Parser`s metadata on its README file
        replace(metadata, contents, local_readme)


@dev.command(help='Adds a few pieces of data to NOMAD.')
@click.option('--username', '-u', type=str, help='The main author username.')
def example_data(username: str):
    from nomad import infrastructure, utils
    from nomad.utils.exampledata import ExampleData

    infrastructure.setup()

    main_author = infrastructure.keycloak.get_user(username=username)
    if main_author is None:
        print(f'The user {username} does not exist.')
        sys.exit(1)

    data = ExampleData(main_author=main_author)

    # one upload with two entries published with embargo, one shared
    upload_id = utils.create_uuid()
    data.create_upload(upload_id=upload_id, published=True, embargo_length=0)
    data.create_entry(
        entry_id=utils.create_uuid(),
        upload_id=upload_id,
        mainfile='test_content/test_embargo_entry/mainfile.json')

    data.save(with_files=True, with_es=True, with_mongo=True)

    return data


@dev.command(help='Creates a Javascript source file containing the required unit conversion factors.')
@click.pass_context
def units(ctx):
    import re
    import json
    from nomad.units import ureg

    # Mapping from unit name to dimension
    unit_map = {
        # Time
        'second': {
            'dimension': 'time',
            'label': 'Second',
            'abbreviation': 's',
        },
        'femtosecond': {
            'dimension': 'time',
            'label': 'Femtosecond',
            'abbreviation': 'fs',
        },
        'atomic_unit_of_time': {
            'dimension': 'time',
            'label': 'Atomic unit of time',
            'abbreviation': 'a_u_time',
        },
        # Length
        'meter': {
            'dimension': 'length',
            'label': 'Meter',
            'abbreviation': 'm',
        },
        'bohr': {
            'dimension': 'length',
            'label': 'Bohr',
            'abbreviation': 'bohr',
        },
        'angstrom': {
            'dimension': 'length',
            'label': 'Ångstrom',
            'abbreviation': 'Å',
        },
        # Mass
        'kilogram': {
            'dimension': 'mass',
            'label': 'Kilogram',
            'abbreviation': 'kg',
        },
        'electron_mass': {
            'dimension': 'mass',
            'label': 'Electron mass',
            'abbreviation': 'mₑ',
        },
        'unified_atomic_mass_unit': {
            'dimension': 'mass',
            'label': 'Unified atomic mass unit',
            'abbreviation': 'u',
        },
        # Current
        'ampere': {
            'dimension': 'current',
            'label': 'Ampere',
            'abbreviation': 'A',
        },
        'atomic_unit_of_current': {
            'dimension': 'current',
            'label': 'Atomic unit of current',
            'abbreviation': 'a_u_current',
        },
        # Substance
        'mole': {
            'dimension': 'substance',
            'label': 'Mole',
            'abbreviation': 'mole',
        },
        # Luminosity
        'candela': {
            'dimension': 'luminosity',
            'label': 'Candela',
            'abbreviation': 'cd',
        },
        # Temperature
        'kelvin': {
            'dimension': 'temperature',
            'label': 'Kelvin',
            'abbreviation': 'K',
        },
        'celsius': {
            'dimension': 'temperature',
            'label': 'Celsius',
            'abbreviation': '°C',
        },
        'fahrenheit': {
            'dimension': 'temperature',
            'label': 'Fahrenheit',
            'abbreviation': '°F',
        },
        'atomic_unit_of_temperature': {
            'dimension': 'temperature',
            'label': 'Atomic unit of temperature',
            'abbreviation': 'a_u_temperature',
        },
        # Force
        'newton': {
            'dimension': 'force',
            'label': 'Newton',
            'abbreviation': 'N',
        },
        'atomic_unit_of_force': {
            'dimension': 'force',
            'label': 'Atomic unit of force',
            'abbreviation': 'a_u_force',
        },
        # Pressure
        'pascal': {
            'dimension': 'pressure',
            'label': 'Pascal',
            'abbreviation': 'Pa'
        },
        'gigapascal': {
            'dimension': 'pressure',
            'label': 'Gigapascal',
            'abbreviation': 'GPa'
        },
        'atomic_unit_of_pressure': {
            'dimension': 'pressure',
            'label': 'Atomic unit of pressure',
            'abbreviation': 'a_u_pressure'
        },
        "millibar": {
            "dimension": "pressure",
            "label": "Millibar",
            "abbreviation": "mbar"
        },
        # Energy
        'joule': {
            'dimension': 'energy',
            'label': 'Joule',
            'abbreviation': 'J',
        },
        'electron_volt': {
            'dimension': 'energy',
            'label': 'Electron volt',
            'abbreviation': 'eV',
        },
        'hartree': {
            'dimension': 'energy',
            'label': 'Hartree',
            'abbreviation': 'Ha',
        },
        # Power
        'watt': {
            'dimension': 'power',
            'label': 'Watt',
            'abbreviation': 'W',
        },
        # Frequency
        'hertz': {
            'dimension': 'frequency',
            'label': 'Hertz',
            'abbreviation': 'Hz',
        },
        # Electric potential
        'volt': {
            'dimension': 'electric_potential',
            'label': 'Volt',
            'abbreviation': 'V',
        },
        # Capacitance
        'farad': {
            'dimension': 'capacitance',
            'label': 'Farad',
            'abbreviation': 'F',
        },
        # Charge
        'coulomb': {
            'dimension': 'charge',
            'label': 'Coulomb',
            'abbreviation': 'C',
        },
        'elementary_charge': {
            'dimension': 'charge',
            'label': 'Elementary charge',
            'abbreviation': 'e',
        },
        # Magnetic field
        'tesla': {
            'dimension': 'magnetic_field',
            'label': 'Tesla',
            'abbreviation': 'T',
        },
        # Magnetic flux
        'weber': {
            'dimension': 'magnetic_flux',
            'label': 'Weber',
            'abbreviation': 'Wb',
        },
        # Magnetic dipole
        'bohr_magneton': {
            'dimension': 'magnetic_dipole',
            'label': 'Bohr magneton',
            'abbreviation': 'Bm',
        },
        # Inductance
        'henry': {
            'dimension': 'inductance',
            'label': 'Henry',
            'abbreviation': 'H',
        },
        # angle
        'radian': {
            'dimension': 'angle',
            'label': 'Radian',
            'abbreviation': 'rad',
        },
        'degree': {
            'dimension': 'angle',
            'label': 'Degree',
            'abbreviation': '°',
        },
        # dimensionless
        'dimensionless': {
            'dimension': 'dimensionless',
            'label': 'Dimensionless',
            'abbreviation': '',
        },
    }

    # Add name as part of the data
    for key, data in unit_map.items():
        data['name'] = key

    # Units that are supported
    unit_table = {
        # Base units
        'time': {
            'units': [
                'second',
                'femtosecond',
                'atomic_unit_of_time',
            ],
            'multipliers': {},
        },
        'length': {
            'units': [
                'meter',
                'bohr',
                'angstrom',
            ],
            'multipliers': {},
        },
        'mass': {
            'units': [
                'kilogram',
                'electron_mass',
                'unified_atomic_mass_unit',
            ],
            'multipliers': {},
        },
        'current': {
            'units': [
                'ampere',
                'atomic_unit_of_current',
            ],
            'multipliers': {},
        },
        'substance': {
            'units': [
                'mole',
            ],
            'multipliers': {},
        },
        'luminosity': {
            'units': [
                'candela',
            ],
            'multipliers': {},
        },
        'temperature': {
            'units': [
                'kelvin',
                'celsius',
                'fahrenheit',
                'atomic_unit_of_temperature',
            ],
            'multipliers': {},
            'constants': {},
        },
        # Derived units
        'force': {
            'units': [
                'newton',
                'atomic_unit_of_force',
            ],
            'multipliers': {},
        },
        "pressure": {
            "units": [
                "pascal",
                "gigapascal",
                "atomic_unit_of_pressure",
                "millibar"
            ],
            'multipliers': {},
        },
        'energy': {
            'units': [
                'joule',
                'electron_volt',
                'hartree',
            ],
            'multipliers': {},
        },
        'power': {
            'units': [
                'watt',
            ],
            'multipliers': {},
        },
        'frequency': {
            'units': [
                'hertz',
            ],
            'multipliers': {},
        },
        'electric_potential': {
            'units': [
                'volt',
            ],
            'multipliers': {},
        },
        'capacitance': {
            'units': [
                'farad',
            ],
            'multipliers': {},
        },
        'charge': {
            'units': [
                'coulomb',
                'elementary_charge',
            ],
            'multipliers': {},
        },
        'magnetic_field': {
            'units': [
                'tesla',
            ],
            'multipliers': {},
        },
        'magnetic_flux': {
            'units': [
                'weber',
            ],
            'multipliers': {},
        },
        'magnetic_dipole': {
            'units': [
                'bohr_magneton',
            ],
            'multipliers': {},
        },
        'inductance': {
            'dimension': 'inductance',
            'units': [
                'henry',
            ],
            'multipliers': {},
        },
        'angle': {
            'dimension': 'angle',
            'units': [
                'radian',
                'degree',
            ],
            'multipliers': {},
        },
        'dimensionless': {
            'dimension': 'dimensionless',
            'units': [
                'dimensionless',
            ],
            'multipliers': {},
        },
    }

    # Unit systems
    unit_systems = {
        'SI': {
            'label': 'SI',
            'description': 'International System of Units (SI)',
            'units': {
                'time': 'second',
                'length': 'meter',
                'mass': 'kilogram',
                'current': 'ampere',
                'substance': 'mole',
                'luminosity': 'candela',
                'temperature': 'kelvin',
                'force': 'newton',
                'pressure': 'pascal',
                'energy': 'joule',
                'power': 'watt',
                'frequency': 'hertz',
                'electric_potential': 'volt',
                'charge': 'coulomb',
                'angle': 'radian',
            },
        },
        'AU': {
            'label': 'Atomic units',
            'description': 'Hartree atomic units',
            'units': {
                'time': 'atomic_unit_of_time',
                'length': 'bohr',
                'mass': 'electron_mass',
                'current': 'atomic_unit_of_current',
                'temperature': 'atomic_unit_of_temperature',
                'force': 'atomic_unit_of_force',
                'energy': 'hartree',
                'pressure': 'atomic_unit_of_pressure',
                'angle': 'radian',
            }
        }
    }

    # Precompute conversion factors and possible shifts
    for value in unit_table.values():
        units = value['units']
        for i_unit in units:
            for j_unit in units:
                if i_unit == j_unit:
                    continue

                # Solve the multiplication factor using 'delta'-units if an
                # offset is present (see
                # https://pint.readthedocs.io/en/0.10.1/nonmult.html) y(1) = a
                # + b -> a = y(1) - b, with delta units b = 0
                y_1 = 1 * getattr(ureg, 'delta_' + i_unit, getattr(ureg, i_unit))
                a = y_1.to(getattr(ureg, 'delta_' + j_unit, getattr(ureg, j_unit))).magnitude

                # Save multiplication factor
                multipliers = value['multipliers']
                if i_unit not in multipliers:
                    multipliers[i_unit] = {}
                if j_unit not in multipliers[i_unit]:
                    multipliers[i_unit][j_unit] = {}
                multipliers[i_unit][j_unit] = a

                # Calculate the constant offset: y = ax + b -> y(0) = b.
                b = ureg.Quantity(0, getattr(ureg, i_unit)).to(getattr(ureg, j_unit)).magnitude

                # Save constant offset is non-zero
                if b != 0:
                    constants = value['constants']
                    if i_unit not in constants:
                        constants[i_unit] = {}
                    if j_unit not in constants[i_unit]:
                        constants[i_unit][j_unit] = {}
                    constants[i_unit][j_unit] = b

    # Check that all defined units are present in the conversion table
    for unit, info in unit_map.items():
        dimension = info['dimension']
        unit_list = unit_table[dimension]['units']
        assert unit in unit_list, 'Could not find "{}" in the unit table under the dimension "{}".'.format(unit, dimension)

    # Check unit system correctness
    for system in unit_systems.values():
        for dimension, unit in system['units'].items():
            info = unit_map[unit]
            assert dimension == info['dimension']

    # Go through the metainfo and check that all units are defined
    all_metainfo = _all_metainfo_packages()
    units = set()
    packages = all_metainfo.m_to_dict(with_meta=True)['packages']
    for package in packages:
        sections = package.get('section_definitions', [])
        for section in sections:
            quantities = section.get('quantities', [])
            for quantity in quantities:
                unit = quantity.get('unit')
                if unit is not None:
                    parts = unit.split()
                    for part in parts:
                        is_operator = part in {'/', '**', '*'}
                        is_number = True
                        try:
                            int(part)
                        except Exception:
                            is_number = False
                        if not is_operator and not is_number:
                            units.add(part)
    for unit in units:
        if unit == 'byte':
            continue

        assert unit in unit_map, 'The unit "{}" is not defined in the unit definitions.'.format(unit)

    # Print unit conversion table and unit systems as a Javascript source file
    output = '''/*
 * Copyright The NOMAD Authors.
 *
 * This file is part of NOMAD. See https://nomad-lab.eu for further info.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
// Generated by NOMAD CLI. Do not edit manually.
'''
    output += 'export const unitMap = '
    json_string = json.dumps(unit_map, indent=2)
    json_string = re.sub(r'(?<!: )"(\S*?)":', '\\1:', json_string)
    json_string = json_string.replace("\"", "'")
    output += json_string
    output += '\nexport const conversionMap = '
    json_string = json.dumps(unit_table, indent=2)
    json_string = re.sub(r'(?<!: )"(\S*?)":', '\\1:', json_string)
    json_string = json_string.replace("\"", "'")
    output += json_string
    output += '\nexport const unitSystems = '
    json_string = json.dumps(unit_systems, indent=2)
    json_string = re.sub(r'(?<!: )"(\S*?)":', '\\1:', json_string)
    json_string = json_string.replace('"', '\'')
    output += json_string
    print(output)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
from nomad.datamodel.metainfo.simulation import run                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, Reference, JSON
)

from nomad.datamodel.metainfo import simulation


m_package = Package()


class x_yambo_io(MSection):
    m_def = Section(validate=False)

    x_yambo_parameters = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_yambo_file = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_yambo_sn = Quantity(
        type=str,
        shape=[],
        description='''
        ''')


class x_yambo_parameters(MSection):
    m_def = Section(validate=False)

    x_yambo_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_k_points = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_g_vectors = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_components = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_symmetries = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_spinor_components = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_spin_polarization = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        ''')

    x_yambo_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_yambo_wf_g_vectors = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_max_atoms_species = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_n_atom_species = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_exact_exchange_fraction_in_xc = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_yambo_exact_exchange_screening_in_xc = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_yambo_magnetic_symmetries = Quantity(
        type=str,
        shape=[],
        description='''
        ''')


class Program(simulation.run.Program):
    m_def = Section(validate=False, extends_base_section=True)

    x_yambo_build = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_yambo_hash = Quantity(
        type=str,
        shape=[],
        description='''
        ''')


class Run(simulation.run.Run):
    m_def = Section(validate=False, extends_base_section=True)

    x_yambo_job_string = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_yambo_cpu = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_threads = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_threads_tot = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_io_nodes = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_fragmented_io = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_yambo_fragmented_wfs = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_yambo_core_databases = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_yambo_additional_io = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_yambo_communications = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_yambo_input_file = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_yambo_report_file = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_yambo_verbose_log_report = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_yambo_precision = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_yambo_log_files = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_yambo_input = SubSection(sub_section=x_yambo_io.m_def, repeats=False)


class x_yambo_local_xc_nonlocal_fock_bandenergies(simulation.calculation.BandEnergies):
    m_def = Section(validate=False)

    x_yambo_sx = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        ''')

    x_yambo_vxc = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        ''')


class x_yambo_bare_xc_bandenergies(simulation.calculation.BandEnergies):
    m_def = Section(validate=False)

    x_yambo_dft = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        ''')

    x_yambo_hf = Quantity(
        type=np.dtype(np.float64),
        shape=['n_spin_channels', 'n_kpoints', 'n_bands'],
        unit='joule',
        description='''
        ''')


class Calculation(simulation.calculation.Calculation):
    m_def = Section(validate=False, extends_base_section=True)

    x_yambo_electronic_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        ''')

    x_yambo_bosonic_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        ''')

    x_yambo_finite_temperature_mode = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_yambo_electronic_density = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        # unit='1/cm**3',
        description='''
        ''')

    x_yambo_filled_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[2],
        description='''
        ''')

    x_yambo_empty_bands = Quantity(
        type=np.dtype(np.int32),
        shape=[2],
        description='''
        ''')

    x_yambo_indirect_gaps = Quantity(
        type=np.dtype(np.float64),
        shape=[2],
        unit='joule',
        description='''
        ''')

    x_yambo_direct_gaps = Quantity(
        type=np.dtype(np.float64),
        shape=[2],
        unit='joule',
        description='''
        ''')

    x_yambo_indirect_gap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        ''')

    x_yambo_direct_gap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        ''')

    x_yambo_direct_gap_k_point = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_indirect_gap_k_points = Quantity(
        type=np.dtype(np.int32),
        shape=[2],
        description='''
        ''')

    x_yambo_local_xc_nonlocal_fock_bandenergies = SubSection(sub_section=x_yambo_local_xc_nonlocal_fock_bandenergies.m_def, repeats=True)

    x_yambo_bare_xc_bandenergies = SubSection(sub_section=x_yambo_bare_xc_bandenergies.m_def, repeats=True)


class x_yambo_dynamic_dielectric_matrix_fragment(MSection):
    m_def = Section(validate=False)

    x_yambo_FREQ_PARS_sec_iq = Quantity(
        type=np.dtype(np.float64),
        shape=[1],
        description='''
        ''')

    x_yambo_FREQ_sec_iq = Quantity(
        type=np.dtype(np.float64),
        shape=[2, 2],
        description='''
        ''')

    x_yambo_matrix_size = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_X_Q = Quantity(
        type=np.dtype(np.float64),
        shape=[2, 'x_yambo_matrix_size', 'x_yambo_matrix_size', 2],
        description='''
        ''')


class x_yambo_module(MSection):
    m_def = Section(validate=False)

    x_yambo_mesh_size = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        ''')

    x_yambo_input = SubSection(sub_section=x_yambo_io.m_def, repeats=True)

    x_yambo_output = SubSection(sub_section=x_yambo_io.m_def, repeats=True)


class x_yambo_dynamic_dielectric_matrix(x_yambo_module):
    m_def = Section(validate=False)

    x_yambo_fragment = SubSection(sub_section=x_yambo_dynamic_dielectric_matrix_fragment.m_def, repeats=True)


class x_yambo_local_xc_nonlocal_fock(x_yambo_module):
    m_def = Section(validate=False)

    x_yambo_plane_waves_vxc = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_yambo_plane_waves_exs = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')


class x_yambo_bare_xc(x_yambo_module):
    m_def = Section(validate=False)

    x_yambo_plane_waves_exs = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')


class x_yambo_dipoles(x_yambo_module):
    m_def = Section(validate=False)


class x_yambo_dyson(x_yambo_module):
    m_def = Section(validate=False)

    x_yambo_bands_range = Quantity(
        type=np.dtype(np.int32),
        shape=[2],
        description='''
        ''')

    x_yambo_g_damping = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')


class x_yambo_transferred_momenta(x_yambo_module):
    m_def = Section(validate=False)


class Method(simulation.method.Method):
    m_def = Section(validate=False, extends_base_section=True)

    x_yambo_transferred_momenta = SubSection(sub_section=x_yambo_transferred_momenta.m_def, repeats=True)

    x_yambo_dynamic_dielectric_matrix = SubSection(sub_section=x_yambo_dynamic_dielectric_matrix.m_def, repeats=True)

    x_yambo_local_xc_nonlocal_fock = SubSection(sub_section=x_yambo_local_xc_nonlocal_fock.m_def, repeats=True)

    x_yambo_bare_xc = SubSection(sub_section=x_yambo_bare_xc.m_def, repeats=True)

    x_yambo_dipoles = SubSection(sub_section=x_yambo_dipoles.m_def, repeats=True)

    x_yambo_dyson = SubSection(sub_section=x_yambo_dyson.m_def, repeats=True)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import logging
import numpy as np
from datetime import datetime

from nomad.units import ureg
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.simulation.run import (
    Run, Program, TimeRun)
from nomad.datamodel.metainfo.simulation.method import (
    Method, DFT, XCFunctional, Functional
)
from nomad.datamodel.metainfo.simulation.system import (
    System, Atoms
)
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, Energy, EnergyEntry, Forces, ForcesEntry, ScfIteration, Charges
)
from .metainfo.onetep import (
    x_onetep_section_tddft, x_onetep_section_tddft_excitations)


re_n = r'[\n\r]'
re_f = r'[-+]?\d+\.\d*(?:[DdEe][-+]\d+)?'


class InputParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def to_block(val_in):
            val = val_in.strip().replace('"', '').splitlines()
            key = val[0].strip()
            return [key.lower(), [v.split() for v in val[1:]]]

        self._quantities = [
            Quantity('parameter', r'([a-z_]+) +: +(\S+)', repeats=True),
            Quantity(
                'block',
                r'\%block +([a-z_]+\s+[\s\S]+?)\s+\%endblock',
                repeats=True, str_operation=to_block)
        ]

    @property
    def cell(self):
        self._results = dict() if self._results is None else self._results
        if self._results.get('cell') is None:

            def resolve_unit(value):
                if isinstance(value[0][0], str):
                    if value[0][0].lower().startswith('ang'):
                        return ureg.angstrom
                    if value[0][0].lower().startswith('bohr'):
                        return ureg.bohr

            cell = dict()
            for block in self.get('block', []):
                if block[0].lower() == 'lattice_cart':
                    unit = resolve_unit(block[1])
                    cell['lattice_vectors'] = block[1] * ureg.bohr if unit is None else block[1][1:] * unit
                elif block[0].lower() == 'positions_abs':
                    unit = resolve_unit(block[1])
                    value = block[1] if unit is None else block[1][1:]
                    unit = ureg.bohr if unit is None else unit
                    cell['labels'] = [v[0] for v in value]
                    cell['positions'] = np.array([v[1:4] for v in value], dtype=np.dtype(np.float64)) * unit
            self._results['cell'] = cell
        return self._results['cell']


class OutParser(TextParser):
    def init_quantities(self):
        calculation_quantities = [
            Quantity(
                'iteration',
                rf'\>\>\> Optimising kernel([\s\S]+?'
                rf'(?:Predicted gain in energy|\-\> RMS NGWF gradient|maximum number of NGWF).+)',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy_total',
                        rf'Total energy += +({re_f}) +Eh',
                        dtype=np.float64, unit=ureg.hartree
                    ),
                    Quantity(
                        'band_gap',
                        rf'Estimated bandgap += +({re_f}) +Eh',
                        dtype=np.float64, unit=ureg.hartree
                    ),
                    Quantity(
                        'x_onetep_rms_occupancy_error',
                        rf'RMS occupancy error += +({re_f})',
                        dtype=np.float64
                    ),
                    Quantity(
                        'x_onetep_commutator',
                        rf'\[H,K\] commutator += +({re_f})',
                        dtype=np.float64
                    ),
                    Quantity(
                        'rms_gradient',
                        rf'RMS gradient += +({re_f})',
                        dtype=np.float64
                    )
                ])
            ),
            Quantity(
                'energy_components',
                r'ENERGY COMPONENTS \(Eh\)([\s\S]+?\={10})',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'contribution', rf'\| (.+): +({re_f})',
                        str_operation=lambda x: [v.strip() for v in x.rsplit(' ', 1)], repeats=True),
                    Quantity('integrated_density', rf'Integrated density +: +({re_f})', dtype=np.float64)
                ])
            ),
            Quantity(
                'forces',
                r'Forces \*+([\s\S]+?TOTAL)',
                sub_parser=TextParser(quantities=[Quantity(
                    'value',
                    rf'[A-Z][a-z]* +\d+ +({re_f} +{re_f} +{re_f})',
                    dtype=np.dtype(np.float64))])
            ),
            Quantity(
                'convergence_tolerance_energy_difference',
                rf'dE\/ion +\| +({re_f}) +\| +({re_f}) +\| +Ha',
                dtype=np.dtype(np.float64), unit=ureg.hartree
            ),
            Quantity(
                'convergence_tolerance_force_maximum',
                rf'\|F\|max +\| +({re_f}) +\| +({re_f}) +\| +Ha\/Bohr',
                dtype=np.dtype(np.float64), unit=ureg.hartree / ureg.bohr
            ),
            Quantity(
                'convergence_tolerance_displacement_maximum',
                rf'\|dR\|max +\| +({re_f}) +\| +({re_f}) +\| +Bohr',
                dtype=np.dtype(np.float64), unit=ureg.bohr
            ),
            Quantity(
                'cell',
                r'Element +Atom +Absolute co\-ordinates of atoms([\s\S]+?)x{10}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'positions',
                        rf'({re_f} +{re_f} +{re_f})',
                        dtype=np.dtype(np.float64), repeats=True),
                    Quantity('labels', r'([A-Z][a-z]*) +\d+', dtype=str, repeats=True)])
            ),
            Quantity(
                'mulliken',
                r'Mulliken Atomic Populations\s+\-+\s+'
                r'Species +Ion +Total +Charge \(e\)\s+\=+\s+([\s\S]+?)\={10}',
                str_operation=lambda x: [v.strip().split() for v in x.strip().splitlines()]
            )
        ]

        self._quantities = [
            Quantity('program_version', r'Version (\S+)', dtype=str),
            Quantity('program_name', r'(\S+) is based on developments described', dtype=str),
            Quantity('x_onetep_number_of_processors', r'Default threads: (\d+)', dtype=np.dtype(np.int32)),
            Quantity('input_file', r'Reading parameters from file \"(.+?)\"', dtype=str),
            Quantity(
                'input',
                r'INPUT FILE \-+([\s\S]+?)\-+ END INPUT FILE',
                sub_parser=TextParser(quantities=[
                    Quantity('parameter', r'([a-z\_]+) +: +(\S+)', repeats=True),
                ])
            ),
            Quantity(
                'date_start',
                r'Job started: (\d\d\-\d\d\-\d\d\d\d \d\d:\d\d \(\+\d\d\d\d\))',
                flatten=False, dtype=str),
            Quantity(
                'date_end',
                r'Job completed: (\d\d\-\d\d\-\d\d\d\d \d\d:\d\d \(\+\d\d\d\d\))',
                flatten=False, dtype=str),
            Quantity(
                'geometry_optimization',
                r'Starting ONETEP Geometry Optimisation([\s\S]+?)(?:\- TIMING|\Z)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'single_point',
                        r'\>\>\> Optimising kernel for current NGWFs:([\s\S]+?\|dR\|max.+)',
                        sub_parser=TextParser(quantities=calculation_quantities)
                    ),
                    Quantity(
                        'iteration',
                        r'(Starting BFGS iteration +\d+[\s\S]+?\|dR\|max.+)',
                        repeats=True, sub_parser=TextParser(quantities=calculation_quantities + [
                            Quantity(
                                'improve',
                                r'BFGS: improving iteration([\s\S]+?\|dR\|max.+)',
                                sub_parser=TextParser(quantities=calculation_quantities)
                            )
                        ])
                    ),
                    Quantity('converged', r'Geometry +optimization +(\S+)', str_operation=lambda x: x == 'completed'),
                    Quantity(
                        'final_configuration',
                        r'(Final Configuration:[\s\S]+?Final Enthalpy.+)',
                        sub_parser=TextParser(quantities=[

                        ])
                    )
                ])
            ),
            Quantity(
                'single_point',
                r'(\>\>\> Optimising kernel for current NGWFs:[\s\S]+?)(?:\- TIMING|\Z)',
                sub_parser=TextParser(quantities=calculation_quantities)
            ),
            Quantity(
                'tddft',
                r'(LR\-TDDFT CG optimisation[\s\S]+?)(?:\- TIMING|\Z)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'iteration',
                        r'(LR_TDDFT CG iteration +\d+ +\#+[\s\S]+?(?:\#{75}|TDDFT_grad))',
                        repeats=True, sub_parser=TextParser(quantities=[
                            Quantity(
                                'energy_total',
                                rf'LR\-TDDFT energy += +({re_f})',
                                dtype=np.float64, unit=ureg.hartree
                            ),
                            Quantity(
                                'x_onetep_tddft_omega_change',
                                rf'Change in omega += +({re_f})',
                                dtype=np.float64
                            ),
                            Quantity(
                                'rms_gradient',
                                rf'RMS gradient += +({re_f})',
                                dtype=np.float64
                            ),
                            Quantity(
                                'x_onetep_tddft_number_conv_states',
                                r'Number of newly converged states: +(\d+)',
                                dtype=np.int32
                            )
                        ])
                    ),
                    Quantity(
                        'excitation',
                        r'\|Excitation\| +Energy \(in Ha\).+([\s\d\.\-\+E]+)',
                        str_operation=lambda x: np.array(
                            [v.strip().split() for v in x.strip().splitlines()],
                            dtype=np.dtype(np.float64))
                    )
                ])
            )
        ]


class OnetepParser:
    def __init__(self):
        self.out_parser = OutParser()
        self.input_parser = InputParser()
        self._xc_map = {
            'PBE': ['GGA_C_PBE', 'GGA_X_PBE'],
            'LDA': ['LDA_C_PZ', 'LDA_X_PZ'],
            'PW91': ['GGA_C_PW91', 'GGA_X_PW91'],
            'PW92': ['GGA_C_PW92', 'GGA_X_PW92'],
            'GGA': ['GGA_X_PBE'],
            'CAPZ': ['LDA_C_PZ'],
            'VWN': ['LDA_C_VWN'],
            'PBESOL': ['GGA_X_PBE_SOL'],
            'RPBE': ['GGA_X_RPBE'],
            'REVPBE': ['GGA_X_PBE_R'],
            'BLYP': ['GGA_X_B88', 'GGA_C_LYP'],
            'XLYP': ['GGA_XC_XLYP'],
            'OPTB88': ['GGA_X_OPTB88_VDW', 'LDA_C_PZ'],
            'OPTPBE': ['GGA_X_OPTPBE_VDW', 'LDA_C_PZ'],
            'VDWDF': ['GGA_X_PBE_R_VDW', 'LDA_C_PZ'],
            'VDWDF2': ['GGA_X_RPW86', 'LDA_C_PZ'],
            'VV10': ['HYB_GGA_XC_LC_VV10'],
            'AVV10S': ['GGA_X_AM05', 'GGA_C_AM05_rVV10-sol']
        }
        self._metainfo_map = {
            'Kinetic': 'kinetic_electronic', 'Hartree': 'correction_hartree',
            'Pseudopotential (local)': 'x_onetep_pseudopotential_local',
            'Pseudopotential (non-local)': 'x_onetep_pseudopotential_non_local',
            'Exchange-correlation': 'xc', 'Ewald': 'ewald', 'Total': 'total'
        }

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger
        self.input_parser.logger = self.logger

    def parse_configuration(self, source):
        def parse_system(source):
            cell = source.get('cell', self.input_parser.cell)
            if cell is None:
                return

            sec_system = self.archive.run[-1].m_create(System)
            positions = cell.get('positions')
            positions = positions * ureg.bohr if isinstance(positions, list) else positions
            sec_system.atoms = Atoms(positions=positions, labels=cell.get('labels'))
            sec_system.atoms.lattice_vectors = cell.get('lattice_vectors', self.input_parser.cell.get('lattice_vectors'))
            return sec_system

        def parse_scc(source):
            sec_scc = self.archive.run[-1].m_create(Calculation)

            # scf iteration
            for iteration in source.get('iteration', []):
                sec_scf = sec_scc.m_create(ScfIteration)
                sec_scf.energy = Energy(
                    total=EnergyEntry(value=iteration.energy_total))

            energies = source.get('energy_components')
            if energies is not None:
                sec_energy = sec_scc.m_create(Energy)
                for contribution in energies.get('contribution', []):
                    key = self._metainfo_map.get(contribution[0])
                    if key is not None:
                        setattr(sec_energy, key, EnergyEntry(value=contribution[1] * ureg.hartree))
            elif len(sec_scc.scf_iteration) > 0:
                sec_scc.energy = Energy(total=EnergyEntry(value=sec_scc.scf_iteration[-1].energy.total.value))

            if source.get('forces') is not None:
                sec_scc.forces = Forces(total=ForcesEntry(value=source.forces.value * (ureg.hartree / ureg.bohr)))

            if source.get('mulliken') is not None:
                sec_charges = sec_scc.m_create(Charges)
                sec_charges.analysis_method = 'Mulliken'
                sec_charges.value = [v[3] for v in source.get('mulliken', [])] * ureg.elementary_charge

            return sec_scc

        sec_system = parse_system(source)
        sec_scc = parse_scc(source)
        sec_scc.system_ref = sec_system

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logging.getLogger(__name__) if logger is None else logger
        self.maindir = os.path.dirname(self.filepath)
        self.init_parser()

        sec_run = self.archive.m_create(Run)
        sec_run.program = Program(
            name=self.out_parser.get('program_name'), version=self.out_parser.get('program_version'))

        def get_timestamp(date):
            if date is None:
                return
            return datetime.strptime(date, '%d-%m-%Y %H:%M (%z)').timestamp()

        sec_run.time_run = TimeRun(
            date_start=get_timestamp(self.out_parser.date_start),
            date_end=get_timestamp(self.out_parser.date_end))

        self.input_parser.mainfile = os.path.join(self.maindir, self.out_parser.get('input_file', ''))

        input_parameters = {key: val for key, val in self.out_parser.get('input', {}).get('parameter', [])}
        sec_method = sec_run.m_create(Method)
        sec_method.x_onetep_input_parameters = input_parameters
        if input_parameters.get('xc_functional') is not None:
            sec_method.dft = DFT(xc_functional=XCFunctional())
            for xc_functional in self._xc_map.get(input_parameters.get('xc_functional'), []):
                if '_X_' in xc_functional or xc_functional.endswith('_X'):
                    sec_method.dft.xc_functional.exchange.append(Functional(name=xc_functional))
                elif '_C_' in xc_functional or xc_functional.endswith('_C'):
                    sec_method.dft.xc_functional.correlation.append(Functional(name=xc_functional))
                elif 'HYB' in xc_functional:
                    sec_method.dft.xc_functional.hybrid.append(Functional(name=xc_functional))
                else:
                    sec_method.dft.xc_functional.contributions.append(Functional(name=xc_functional))

        if self.out_parser.geometry_optimization is not None:
            if self.out_parser.geometry_optimization.single_point is not None:
                self.parse_configuration(self.out_parser.geometry_optimization.single_point)
            for iteration in self.out_parser.geometry_optimization.get('iteration', []):
                self.parse_configuration(iteration if iteration.improve is None else iteration.improve)

        elif self.out_parser.single_point is not None:
            self.parse_configuration(self.out_parser.single_point)

        elif self.out_parser.tddft is not None:
            self.parse_configuration(self.out_parser.tddft)
            if self.out_parser.tddft is not None:
                sec_tddft = sec_run.calculation[-1].m_create(x_onetep_section_tddft)
                for excitation in self.out_parser.tddft.get('excitation', []):
                    sec_excitation = sec_tddft.m_create(x_onetep_section_tddft_excitations)
                    sec_excitation.x_onetep_tddft_excit_energy = excitation[1] * ureg.hartree
                    sec_excitation.x_onetep_tddft_excit_oscill_str = excitation[2]
                    sec_excitation.x_onetep_tddft_excit_lifetime = excitation[3] * ureg.ns
        # TODO parse md
